"""Comprehensive tests for distributed tracing module.

Tests cover:
- Span creation and lifecycle
- TracerProvider and Tracer
- Span processors (Simple, Batch, Multi)
- Span exporters (Console, InMemory, OTLP, Jaeger, Zipkin)
- Samplers (AlwaysOn, AlwaysOff, TraceIdRatio, ParentBased)
- Propagators (W3C, B3, Jaeger)
- Resource detection
- Baggage management
- Configuration utilities
"""

from __future__ import annotations

import json
import os
import threading
import time
from io import StringIO
from unittest.mock import MagicMock, patch

import pytest

from truthound.observability.tracing import (
    # Provider
    TracerProvider,
    Tracer,
    get_tracer_provider,
    set_tracer_provider,
    # Span
    Span,
    SpanKind,
    StatusCode,
    Link,
    Event,
    SpanLimits,
    # Processor
    SpanProcessor,
    SimpleSpanProcessor,
    BatchSpanProcessor,
    MultiSpanProcessor,
    # Exporter
    SpanExporter,
    ExportResult,
    ConsoleSpanExporter,
    InMemorySpanExporter,
    OTLPSpanExporter,
    JaegerExporter,
    ZipkinExporter,
    # Sampler
    Sampler,
    SamplingResult,
    SamplingDecision,
    AlwaysOnSampler,
    AlwaysOffSampler,
    TraceIdRatioSampler,
    ParentBasedSampler,
    # Propagator
    Propagator,
    CompositePropagator,
    W3CTraceContextPropagator,
    W3CBaggagePropagator,
    B3Propagator,
    JaegerPropagator,
    get_global_propagator,
    set_global_propagator,
    # Resource
    Resource,
    ResourceDetector,
    get_aggregated_resources,
    # Baggage
    Baggage,
    get_baggage,
    set_baggage,
    remove_baggage,
    clear_baggage,
    # Config
    TracingConfig,
    configure_tracing,
)

from truthound.observability.tracing.span import (
    SpanContextData,
    generate_trace_id,
    generate_span_id,
    NoOpSpan,
)

from truthound.observability.tracing.provider import (
    get_tracer,
    get_current_span,
)

from truthound.observability.tracing.resource import (
    ProcessResourceDetector,
    HostResourceDetector,
    ServiceResourceDetector,
    SDKResourceDetector,
)

from truthound.observability.tracing.baggage import baggage_context

from truthound.observability.tracing.propagator import inject_context, extract_context


# =============================================================================
# Span Tests
# =============================================================================


class TestSpanContextData:
    """Tests for SpanContextData."""

    def test_creation(self):
        """Test span context creation."""
        ctx = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
        )
        assert ctx.is_valid
        assert ctx.is_sampled  # Default trace_flags=1

    def test_not_sampled(self):
        """Test unsampled span context."""
        ctx = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
            trace_flags=0,
        )
        assert ctx.is_valid
        assert not ctx.is_sampled

    def test_to_w3c_traceparent(self):
        """Test W3C traceparent generation."""
        ctx = SpanContextData(
            trace_id="0af7651916cd43dd8448eb211c80319c",
            span_id="b7ad6b7169203331",
            trace_flags=1,
        )
        traceparent = ctx.to_w3c_traceparent()
        assert traceparent == "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"

    def test_from_w3c_traceparent(self):
        """Test W3C traceparent parsing."""
        header = "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"
        ctx = SpanContextData.from_w3c_traceparent(header)

        assert ctx is not None
        assert ctx.trace_id == "0af7651916cd43dd8448eb211c80319c"
        assert ctx.span_id == "b7ad6b7169203331"
        assert ctx.is_sampled
        assert ctx.is_remote


class TestSpan:
    """Tests for Span."""

    def test_creation(self):
        """Test span creation."""
        ctx = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
        )
        span = Span(name="test-operation", context=ctx)

        assert span.name == "test-operation"
        assert span.context == ctx
        assert span.is_recording()
        assert span.kind == SpanKind.INTERNAL

    def test_attributes(self):
        """Test setting span attributes."""
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )

        span.set_attribute("key1", "value1")
        span.set_attribute("key2", 42)
        span.set_attributes({"key3": True, "key4": 3.14})

        assert span.attributes["key1"] == "value1"
        assert span.attributes["key2"] == 42
        assert span.attributes["key3"] is True
        assert span.attributes["key4"] == 3.14

    def test_events(self):
        """Test adding span events."""
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )

        span.add_event("event1", {"attr": "value"})
        span.add_event("event2")

        assert len(span.events) == 2
        assert span.events[0].name == "event1"
        assert span.events[0].attributes["attr"] == "value"
        assert span.events[1].name == "event2"

    def test_status(self):
        """Test setting span status."""
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )

        span.set_status(StatusCode.OK)
        assert span.status == (StatusCode.OK, "")

        # Once ERROR is set, it cannot be changed
        span.set_status(StatusCode.ERROR, "Something went wrong")
        assert span.status == (StatusCode.ERROR, "Something went wrong")

        span.set_status(StatusCode.OK)  # Should be ignored
        assert span.status[0] == StatusCode.ERROR

    def test_end(self):
        """Test ending a span."""
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )

        assert span.is_recording()
        assert span.end_time is None

        time.sleep(0.01)
        span.end()

        assert not span.is_recording()
        assert span.end_time is not None
        assert span.duration_ms >= 10

    def test_context_manager(self):
        """Test span as context manager."""
        ctx = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
        )
        span = Span(name="test", context=ctx)

        with span as s:
            s.set_attribute("in_context", True)

        assert not span.is_recording()
        assert span.status[0] == StatusCode.OK

    def test_context_manager_exception(self):
        """Test span context manager with exception."""
        ctx = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
        )
        span = Span(name="test", context=ctx)

        with pytest.raises(ValueError):
            with span:
                raise ValueError("Test error")

        assert not span.is_recording()
        assert span.status[0] == StatusCode.ERROR
        assert "Test error" in span.status[1]

    def test_record_exception(self):
        """Test recording exception."""
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )

        try:
            raise ValueError("Test exception")
        except ValueError as e:
            span.record_exception(e)

        assert len(span.events) == 1
        assert span.events[0].name == "exception"
        assert span.events[0].attributes["exception.type"] == "ValueError"

    def test_span_limits(self):
        """Test span limits."""
        limits = SpanLimits(max_attributes=2, max_events=2)
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
            limits=limits,
        )

        span.set_attribute("a", 1)
        span.set_attribute("b", 2)
        span.set_attribute("c", 3)  # Should be ignored

        assert len(span.attributes) == 2
        assert "c" not in span.attributes

    def test_to_dict(self):
        """Test span serialization."""
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id="abcd1234" * 4,
                span_id="efgh5678" * 2,
            ),
            kind=SpanKind.SERVER,
        )
        span.set_attribute("key", "value")
        span.add_event("event")
        span.end()

        data = span.to_dict()

        assert data["name"] == "test"
        assert data["kind"] == "SERVER"
        assert data["attributes"]["key"] == "value"
        assert len(data["events"]) == 1


class TestNoOpSpan:
    """Tests for NoOpSpan."""

    def test_is_no_op(self):
        """Test NoOpSpan doesn't record."""
        span = NoOpSpan()

        assert not span.is_recording()
        span.set_attribute("key", "value")  # Should be no-op
        span.add_event("event")  # Should be no-op

    def test_context_manager(self):
        """Test NoOpSpan as context manager."""
        span = NoOpSpan()

        with span as s:
            pass

        # Should not raise


# =============================================================================
# Provider Tests
# =============================================================================


class TestTracerProvider:
    """Tests for TracerProvider."""

    def test_get_tracer(self):
        """Test getting a tracer."""
        provider = TracerProvider()
        tracer = provider.get_tracer("my.component", "1.0.0")

        assert isinstance(tracer, Tracer)
        assert tracer.name == "my.component"
        assert tracer.version == "1.0.0"

    def test_tracer_caching(self):
        """Test tracer caching."""
        provider = TracerProvider()

        tracer1 = provider.get_tracer("component", "1.0")
        tracer2 = provider.get_tracer("component", "1.0")

        assert tracer1 is tracer2

    def test_start_span(self):
        """Test starting spans."""
        exporter = InMemorySpanExporter()
        processor = SimpleSpanProcessor(exporter)
        provider = TracerProvider(processors=[processor])

        tracer = provider.get_tracer("test")
        with tracer.start_as_current_span("operation") as span:
            span.set_attribute("key", "value")

        spans = exporter.get_finished_spans()
        assert len(spans) == 1
        assert spans[0].name == "operation"

    def test_nested_spans(self):
        """Test nested spans."""
        exporter = InMemorySpanExporter()
        processor = SimpleSpanProcessor(exporter)
        provider = TracerProvider(processors=[processor])

        tracer = provider.get_tracer("test")

        with tracer.start_as_current_span("parent") as parent:
            with tracer.start_as_current_span("child") as child:
                assert child.parent is not None
                assert child.parent.span_id == parent.context.span_id

        spans = exporter.get_finished_spans()
        assert len(spans) == 2


class TestGlobalProvider:
    """Tests for global provider functions."""

    def test_set_get_provider(self):
        """Test setting and getting global provider."""
        provider = TracerProvider()
        set_tracer_provider(provider)

        assert get_tracer_provider() is provider

    def test_get_tracer_convenience(self):
        """Test get_tracer convenience function."""
        provider = TracerProvider()
        set_tracer_provider(provider)

        tracer = get_tracer("test.component")
        assert isinstance(tracer, Tracer)


# =============================================================================
# Sampler Tests
# =============================================================================


class TestSamplers:
    """Tests for samplers."""

    def test_always_on(self):
        """Test AlwaysOnSampler."""
        sampler = AlwaysOnSampler()
        result = sampler.should_sample(
            parent_context=None,
            trace_id=generate_trace_id(),
            name="test",
            kind=SpanKind.INTERNAL,
        )

        assert result.is_sampled
        assert result.decision == SamplingDecision.RECORD_AND_SAMPLE

    def test_always_off(self):
        """Test AlwaysOffSampler."""
        sampler = AlwaysOffSampler()
        result = sampler.should_sample(
            parent_context=None,
            trace_id=generate_trace_id(),
            name="test",
            kind=SpanKind.INTERNAL,
        )

        assert not result.is_sampled
        assert result.decision == SamplingDecision.DROP

    def test_trace_id_ratio(self):
        """Test TraceIdRatioSampler."""
        sampler = TraceIdRatioSampler(0.5)

        sampled = 0
        total = 1000

        for _ in range(total):
            result = sampler.should_sample(
                parent_context=None,
                trace_id=generate_trace_id(),
                name="test",
                kind=SpanKind.INTERNAL,
            )
            if result.is_sampled:
                sampled += 1

        # Should be roughly 50%
        ratio = sampled / total
        assert 0.4 < ratio < 0.6

    def test_trace_id_ratio_deterministic(self):
        """Test TraceIdRatioSampler is deterministic."""
        sampler = TraceIdRatioSampler(0.5)
        trace_id = generate_trace_id()

        result1 = sampler.should_sample(
            None, trace_id, "test", SpanKind.INTERNAL
        )
        result2 = sampler.should_sample(
            None, trace_id, "test", SpanKind.INTERNAL
        )

        assert result1.is_sampled == result2.is_sampled

    def test_parent_based(self):
        """Test ParentBasedSampler."""
        sampler = ParentBasedSampler(root=TraceIdRatioSampler(1.0))

        # Root span
        result = sampler.should_sample(
            parent_context=None,
            trace_id=generate_trace_id(),
            name="test",
            kind=SpanKind.INTERNAL,
        )
        assert result.is_sampled

        # With sampled parent
        parent = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
            trace_flags=1,
        )
        result = sampler.should_sample(
            parent_context=parent,
            trace_id=parent.trace_id,
            name="test",
            kind=SpanKind.INTERNAL,
        )
        assert result.is_sampled


# =============================================================================
# Processor Tests
# =============================================================================


class TestSimpleSpanProcessor:
    """Tests for SimpleSpanProcessor."""

    def test_export_on_end(self):
        """Test spans are exported on end."""
        exporter = InMemorySpanExporter()
        processor = SimpleSpanProcessor(exporter)

        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )

        processor.on_start(span)
        span.end()
        processor.on_end(span)

        assert len(exporter.get_finished_spans()) == 1

    def test_shutdown(self):
        """Test processor shutdown."""
        exporter = InMemorySpanExporter()
        processor = SimpleSpanProcessor(exporter)

        assert processor.shutdown()


class TestBatchSpanProcessor:
    """Tests for BatchSpanProcessor."""

    def test_batch_export(self):
        """Test batch export."""
        from truthound.observability.tracing.processor import BatchConfig

        exporter = InMemorySpanExporter()
        processor = BatchSpanProcessor(
            exporter,
            config=BatchConfig(
                scheduled_delay_millis=100,
                max_export_batch_size=10,
            ),
        )

        # Add spans
        for i in range(5):
            span = Span(
                name=f"test-{i}",
                context=SpanContextData(
                    trace_id=generate_trace_id(),
                    span_id=generate_span_id(),
                ),
            )
            span.end()
            processor.on_end(span)

        # Force flush
        processor.force_flush()

        # Check exported
        assert len(exporter.get_finished_spans()) == 5

        processor.shutdown()

    def test_shutdown_exports_pending(self):
        """Test shutdown exports pending spans."""
        exporter = InMemorySpanExporter()
        processor = BatchSpanProcessor(exporter)

        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )
        span.end()
        processor.on_end(span)

        processor.shutdown()

        assert len(exporter.get_finished_spans()) >= 1


class TestMultiSpanProcessor:
    """Tests for MultiSpanProcessor."""

    def test_fan_out(self):
        """Test fan-out to multiple processors."""
        exporter1 = InMemorySpanExporter()
        exporter2 = InMemorySpanExporter()

        multi = MultiSpanProcessor([
            SimpleSpanProcessor(exporter1),
            SimpleSpanProcessor(exporter2),
        ])

        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )
        span.end()

        multi.on_end(span)

        assert len(exporter1.get_finished_spans()) == 1
        assert len(exporter2.get_finished_spans()) == 1


# =============================================================================
# Exporter Tests
# =============================================================================


class TestConsoleSpanExporter:
    """Tests for ConsoleSpanExporter."""

    def test_export_pretty(self):
        """Test pretty print export."""
        output = StringIO()
        exporter = ConsoleSpanExporter(output=output, pretty=True)

        span = Span(
            name="test-operation",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )
        span.set_attribute("key", "value")
        span.end()

        result = exporter.export([span])

        assert result == ExportResult.SUCCESS
        output_str = output.getvalue()
        assert "test-operation" in output_str
        assert "key" in output_str

    def test_export_json(self):
        """Test JSON export."""
        output = StringIO()
        exporter = ConsoleSpanExporter(output=output, json_output=True)

        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )
        span.end()

        exporter.export([span])

        output_str = output.getvalue()
        data = json.loads(output_str.strip())
        assert data["name"] == "test"


class TestInMemorySpanExporter:
    """Tests for InMemorySpanExporter."""

    def test_export_and_get(self):
        """Test export and retrieval."""
        exporter = InMemorySpanExporter()

        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )
        span.end()

        exporter.export([span])

        spans = exporter.get_finished_spans()
        assert len(spans) == 1
        assert spans[0].name == "test"

    def test_clear(self):
        """Test clearing spans."""
        exporter = InMemorySpanExporter()

        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )
        span.end()

        exporter.export([span])
        exporter.clear()

        assert len(exporter.get_finished_spans()) == 0


# =============================================================================
# Propagator Tests
# =============================================================================


class TestW3CTraceContextPropagator:
    """Tests for W3C Trace Context propagator."""

    def test_inject(self):
        """Test context injection."""
        propagator = W3CTraceContextPropagator()
        ctx = SpanContextData(
            trace_id="0af7651916cd43dd8448eb211c80319c",
            span_id="b7ad6b7169203331",
            trace_flags=1,
        )

        headers = {}
        propagator.inject(ctx, headers)

        assert "traceparent" in headers
        assert headers["traceparent"] == "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"

    def test_extract(self):
        """Test context extraction."""
        propagator = W3CTraceContextPropagator()
        headers = {
            "traceparent": "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01",
        }

        ctx = propagator.extract(headers)

        assert ctx is not None
        assert ctx.trace_id == "0af7651916cd43dd8448eb211c80319c"
        assert ctx.span_id == "b7ad6b7169203331"
        assert ctx.is_sampled

    def test_roundtrip(self):
        """Test inject/extract roundtrip."""
        propagator = W3CTraceContextPropagator()
        original = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
            trace_flags=1,
        )

        headers = {}
        propagator.inject(original, headers)
        extracted = propagator.extract(headers)

        assert extracted is not None
        assert extracted.trace_id == original.trace_id
        assert extracted.span_id == original.span_id


class TestB3Propagator:
    """Tests for B3 propagator."""

    def test_inject_multi_header(self):
        """Test multi-header injection."""
        propagator = B3Propagator(single_header=False)
        ctx = SpanContextData(
            trace_id="0af7651916cd43dd8448eb211c80319c",
            span_id="b7ad6b7169203331",
            trace_flags=1,
        )

        headers = {}
        propagator.inject(ctx, headers)

        assert headers["X-B3-TraceId"] == "0af7651916cd43dd8448eb211c80319c"
        assert headers["X-B3-SpanId"] == "b7ad6b7169203331"
        assert headers["X-B3-Sampled"] == "1"

    def test_inject_single_header(self):
        """Test single-header injection."""
        propagator = B3Propagator(single_header=True)
        ctx = SpanContextData(
            trace_id="0af7651916cd43dd8448eb211c80319c",
            span_id="b7ad6b7169203331",
            trace_flags=1,
        )

        headers = {}
        propagator.inject(ctx, headers)

        assert "b3" in headers
        assert headers["b3"] == "0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-1"

    def test_extract_multi_header(self):
        """Test multi-header extraction."""
        propagator = B3Propagator()
        headers = {
            "X-B3-TraceId": "0af7651916cd43dd8448eb211c80319c",
            "X-B3-SpanId": "b7ad6b7169203331",
            "X-B3-Sampled": "1",
        }

        ctx = propagator.extract(headers)

        assert ctx is not None
        assert ctx.trace_id == "0af7651916cd43dd8448eb211c80319c"
        assert ctx.is_sampled


class TestJaegerPropagator:
    """Tests for Jaeger propagator."""

    def test_inject(self):
        """Test Jaeger header injection."""
        propagator = JaegerPropagator()
        ctx = SpanContextData(
            trace_id="0af7651916cd43dd8448eb211c80319c",
            span_id="b7ad6b7169203331",
            trace_flags=1,
        )

        headers = {}
        propagator.inject(ctx, headers)

        assert "uber-trace-id" in headers
        assert "0af7651916cd43dd8448eb211c80319c" in headers["uber-trace-id"]

    def test_extract(self):
        """Test Jaeger header extraction."""
        propagator = JaegerPropagator()
        headers = {
            "uber-trace-id": "0af7651916cd43dd8448eb211c80319c:b7ad6b7169203331:0:1",
        }

        ctx = propagator.extract(headers)

        assert ctx is not None
        assert ctx.trace_id == "0af7651916cd43dd8448eb211c80319c"


class TestCompositePropagator:
    """Tests for CompositePropagator."""

    def test_inject_all(self):
        """Test injection to all propagators."""
        composite = CompositePropagator([
            W3CTraceContextPropagator(),
            B3Propagator(),
        ])

        ctx = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
        )

        headers = {}
        composite.inject(ctx, headers)

        assert "traceparent" in headers
        assert "X-B3-TraceId" in headers

    def test_extract_first_match(self):
        """Test extraction from first matching propagator."""
        composite = CompositePropagator([
            W3CTraceContextPropagator(),
            B3Propagator(),
        ])

        headers = {
            "traceparent": "00-" + "a" * 32 + "-" + "b" * 16 + "-01",
        }

        ctx = composite.extract(headers)
        assert ctx is not None


# =============================================================================
# Resource Tests
# =============================================================================


class TestResource:
    """Tests for Resource."""

    def test_creation(self):
        """Test resource creation."""
        resource = Resource(attributes={
            "service.name": "test-service",
            "service.version": "1.0.0",
        })

        assert resource.get("service.name") == "test-service"
        assert resource.get("service.version") == "1.0.0"

    def test_merge(self):
        """Test resource merging."""
        r1 = Resource(attributes={"a": 1, "b": 2})
        r2 = Resource(attributes={"b": 3, "c": 4})

        merged = r1.merge(r2)

        assert merged.get("a") == 1
        assert merged.get("b") == 3  # Overwritten
        assert merged.get("c") == 4


class TestResourceDetectors:
    """Tests for resource detectors."""

    def test_process_detector(self):
        """Test ProcessResourceDetector."""
        detector = ProcessResourceDetector()
        resource = detector.detect()

        assert resource.get("process.pid") is not None
        assert resource.get("process.runtime.name") is not None

    def test_host_detector(self):
        """Test HostResourceDetector."""
        detector = HostResourceDetector()
        resource = detector.detect()

        assert resource.get("host.name") is not None

    def test_service_detector(self):
        """Test ServiceResourceDetector."""
        detector = ServiceResourceDetector(service_name="test-service")
        resource = detector.detect()

        assert resource.get("service.name") == "test-service"

    def test_sdk_detector(self):
        """Test SDKResourceDetector."""
        detector = SDKResourceDetector()
        resource = detector.detect()

        assert resource.get("telemetry.sdk.name") == "truthound"

    def test_aggregated_resources(self):
        """Test get_aggregated_resources."""
        resource = get_aggregated_resources()

        # Should have attributes from multiple detectors
        assert resource.get("telemetry.sdk.name") is not None
        assert resource.get("process.pid") is not None


# =============================================================================
# Baggage Tests
# =============================================================================


class TestBaggage:
    """Tests for Baggage."""

    def test_set_get(self):
        """Test setting and getting baggage."""
        baggage = Baggage()
        baggage = baggage.set("key", "value")

        assert baggage.get("key") == "value"

    def test_remove(self):
        """Test removing baggage."""
        baggage = Baggage().set("key", "value")
        baggage = baggage.remove("key")

        assert baggage.get("key") is None

    def test_immutability(self):
        """Test baggage immutability."""
        b1 = Baggage()
        b2 = b1.set("key", "value")

        assert b1.get("key") is None
        assert b2.get("key") == "value"


class TestBaggageContext:
    """Tests for baggage context management."""

    def test_context_functions(self):
        """Test baggage context functions."""
        clear_baggage()

        set_baggage("user_id", "123")
        assert get_baggage().get("user_id") == "123"

        remove_baggage("user_id")
        assert get_baggage().get("user_id") is None

    def test_context_manager(self):
        """Test baggage_context context manager."""
        clear_baggage()

        with baggage_context(user_id="123", tenant="acme"):
            assert get_baggage().get("user_id") == "123"
            assert get_baggage().get("tenant") == "acme"

        # Should be restored
        assert get_baggage().get("user_id") is None


# =============================================================================
# Configuration Tests
# =============================================================================


class TestTracingConfig:
    """Tests for TracingConfig."""

    def test_default_config(self):
        """Test default configuration."""
        config = TracingConfig()

        assert config.service_name == "unknown_service"
        assert config.exporter == "console"
        assert config.sampler == "always_on"

    def test_development_config(self):
        """Test development configuration."""
        config = TracingConfig.development("dev-service")

        assert config.service_name == "dev-service"
        assert config.exporter == "console"
        assert config.batch_export is False

    def test_production_config(self):
        """Test production configuration."""
        config = TracingConfig.production(
            service_name="prod-service",
            endpoint="http://collector:4317",
            sampling_ratio=0.1,
        )

        assert config.service_name == "prod-service"
        assert config.exporter == "otlp"
        assert config.sampling_ratio == 0.1

    def test_from_env(self):
        """Test configuration from environment."""
        with patch.dict(os.environ, {
            "OTEL_SERVICE_NAME": "env-service",
            "OTEL_SERVICE_VERSION": "1.2.3",
        }):
            config = TracingConfig.from_env()

        assert config.service_name == "env-service"
        assert config.service_version == "1.2.3"


class TestConfigureTracing:
    """Tests for configure_tracing."""

    def test_basic_configuration(self):
        """Test basic tracing configuration."""
        provider = configure_tracing(
            service_name="test-service",
            exporter="console",
            set_global=False,
        )

        assert isinstance(provider, TracerProvider)
        assert provider.resource.get("service.name") == "test-service"

    def test_with_config_object(self):
        """Test configuration with TracingConfig."""
        config = TracingConfig(
            service_name="config-service",
            exporter="console",
            sampler="always_on",
        )

        provider = configure_tracing(config, set_global=False)

        assert provider.resource.get("service.name") == "config-service"


# =============================================================================
# Thread Safety Tests
# =============================================================================


class TestThreadSafety:
    """Tests for thread safety."""

    def test_span_thread_safety(self):
        """Test span operations are thread-safe."""
        span = Span(
            name="test",
            context=SpanContextData(
                trace_id=generate_trace_id(),
                span_id=generate_span_id(),
            ),
        )

        def add_attributes():
            for i in range(100):
                span.set_attribute(f"key_{threading.current_thread().name}_{i}", i)

        threads = [threading.Thread(target=add_attributes) for _ in range(10)]

        for t in threads:
            t.start()
        for t in threads:
            t.join()

        # Should have attributes from all threads
        assert len(span.attributes) > 0

    def test_exporter_thread_safety(self):
        """Test exporter is thread-safe."""
        exporter = InMemorySpanExporter()

        def export_span():
            for _ in range(100):
                span = Span(
                    name="test",
                    context=SpanContextData(
                        trace_id=generate_trace_id(),
                        span_id=generate_span_id(),
                    ),
                )
                span.end()
                exporter.export([span])

        threads = [threading.Thread(target=export_span) for _ in range(5)]

        for t in threads:
            t.start()
        for t in threads:
            t.join()

        assert len(exporter.get_finished_spans()) == 500


# =============================================================================
# Integration Tests
# =============================================================================


class TestIntegration:
    """Integration tests for complete tracing workflows."""

    def test_end_to_end_tracing(self):
        """Test complete tracing workflow."""
        exporter = InMemorySpanExporter()
        processor = SimpleSpanProcessor(exporter)
        provider = TracerProvider(
            resource=Resource(attributes={"service.name": "integration-test"}),
            processors=[processor],
        )

        tracer = provider.get_tracer("test.integration", "1.0.0")

        with tracer.start_as_current_span("parent-operation") as parent:
            parent.set_attribute("parent.attr", "value")

            with tracer.start_as_current_span("child-operation") as child:
                child.add_event("processing", {"step": 1})

        spans = exporter.get_finished_spans()
        assert len(spans) == 2

        # Check parent-child relationship
        child_span = [s for s in spans if s.name == "child-operation"][0]
        parent_span = [s for s in spans if s.name == "parent-operation"][0]

        assert child_span.parent.span_id == parent_span.context.span_id

    def test_context_propagation(self):
        """Test context propagation across boundaries."""
        # Inject context
        ctx = SpanContextData(
            trace_id=generate_trace_id(),
            span_id=generate_span_id(),
            trace_flags=1,
        )

        headers = {}
        inject_context(ctx, headers)

        # Extract and continue trace
        extracted = extract_context(headers)

        assert extracted is not None
        assert extracted.trace_id == ctx.trace_id
        assert extracted.is_remote


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Lock manager for coordinating file access.

This module provides high-level lock management with features like:
- Deadlock detection
- Lock timeout handling
- Reentrant locks
- Lock statistics and debugging

The FileLockManager is the main entry point for most use cases.
"""

from __future__ import annotations

import threading
import time
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Iterator

from truthound.stores.concurrency.locks import (
    LockHandle,
    LockMode,
    LockStrategy,
    get_default_lock_strategy,
)


class LockTimeout(Exception):
    """Raised when lock acquisition times out."""

    def __init__(self, path: Path, timeout: float) -> None:
        self.path = path
        self.timeout = timeout
        super().__init__(f"Timeout acquiring lock on {path} after {timeout}s")


class DeadlockError(Exception):
    """Raised when a potential deadlock is detected."""

    def __init__(self, path: Path, held_locks: list[Path]) -> None:
        self.path = path
        self.held_locks = held_locks
        super().__init__(
            f"Potential deadlock: trying to acquire {path} while holding {held_locks}"
        )


@dataclass
class LockStatistics:
    """Statistics about lock usage.

    Useful for debugging and performance monitoring.
    """

    total_acquisitions: int = 0
    total_releases: int = 0
    total_wait_time: float = 0.0
    total_hold_time: float = 0.0
    max_wait_time: float = 0.0
    max_hold_time: float = 0.0
    contentions: int = 0
    timeouts: int = 0
    deadlocks: int = 0

    def record_acquisition(self, wait_time: float, contended: bool) -> None:
        """Record a lock acquisition."""
        self.total_acquisitions += 1
        self.total_wait_time += wait_time
        self.max_wait_time = max(self.max_wait_time, wait_time)
        if contended:
            self.contentions += 1

    def record_release(self, hold_time: float) -> None:
        """Record a lock release."""
        self.total_releases += 1
        self.total_hold_time += hold_time
        self.max_hold_time = max(self.max_hold_time, hold_time)

    def record_timeout(self) -> None:
        """Record a timeout."""
        self.timeouts += 1

    def record_deadlock(self) -> None:
        """Record a deadlock detection."""
        self.deadlocks += 1

    @property
    def avg_wait_time(self) -> float:
        """Average wait time for lock acquisition."""
        if self.total_acquisitions == 0:
            return 0.0
        return self.total_wait_time / self.total_acquisitions

    @property
    def avg_hold_time(self) -> float:
        """Average time locks are held."""
        if self.total_releases == 0:
            return 0.0
        return self.total_hold_time / self.total_releases

    @property
    def contention_rate(self) -> float:
        """Percentage of acquisitions that were contended."""
        if self.total_acquisitions == 0:
            return 0.0
        return self.contentions / self.total_acquisitions * 100


@dataclass
class LockContext:
    """Context for a held lock.

    Tracks information about the lock for debugging and management.
    """

    handle: LockHandle
    acquired_at: float = field(default_factory=time.time)
    thread_id: int = field(default_factory=threading.get_ident)
    reentry_count: int = 1
    wait_time: float = 0.0

    @property
    def hold_time(self) -> float:
        """How long the lock has been held."""
        return time.time() - self.acquired_at


class FileLockManager:
    """High-level manager for file locking.

    Provides a thread-safe interface for acquiring and releasing locks
    with support for:
    - Multiple locking strategies
    - Reentrant locks (same thread can acquire multiple times)
    - Deadlock detection
    - Lock statistics
    - Automatic cleanup

    Example:
        >>> manager = FileLockManager()
        >>>
        >>> # Using context manager
        >>> with manager.acquire(Path("/path/to/file"), LockMode.EXCLUSIVE):
        ...     # File is locked
        ...     pass
        >>>
        >>> # Manual acquire/release
        >>> handle = manager.lock(Path("/path/to/file"), LockMode.SHARED)
        >>> try:
        ...     # Do work
        ...     pass
        ... finally:
        ...     manager.unlock(handle)
    """

    def __init__(
        self,
        strategy: LockStrategy | None = None,
        enable_deadlock_detection: bool = True,
        enable_statistics: bool = True,
        default_timeout: float | None = 30.0,
    ) -> None:
        """Initialize the lock manager.

        Args:
            strategy: Lock strategy to use. If None, auto-detects best strategy.
            enable_deadlock_detection: Whether to detect potential deadlocks.
            enable_statistics: Whether to collect lock statistics.
            default_timeout: Default timeout for lock acquisition (None = infinite).
        """
        self._strategy = strategy or get_default_lock_strategy()
        self._enable_deadlock_detection = enable_deadlock_detection
        self._enable_statistics = enable_statistics
        self._default_timeout = default_timeout

        # Thread-local storage for tracking held locks
        self._thread_locks: threading.local = threading.local()

        # Global lock tracking
        self._global_lock = threading.RLock()
        self._held_locks: dict[str, LockContext] = {}

        # Statistics
        self._stats = LockStatistics() if enable_statistics else None
        self._path_stats: dict[str, LockStatistics] = defaultdict(LockStatistics)

    @property
    def statistics(self) -> LockStatistics | None:
        """Get global lock statistics."""
        return self._stats

    def get_path_statistics(self, path: Path) -> LockStatistics:
        """Get statistics for a specific path."""
        return self._path_stats[str(path)]

    def _get_thread_locks(self) -> dict[str, LockContext]:
        """Get locks held by the current thread."""
        if not hasattr(self._thread_locks, "locks"):
            self._thread_locks.locks = {}
        return self._thread_locks.locks

    def _check_deadlock(self, path: Path) -> None:
        """Check for potential deadlock.

        Simple cycle detection: if thread already holds locks and is
        trying to acquire a new one, there's potential for deadlock.
        """
        if not self._enable_deadlock_detection:
            return

        thread_locks = self._get_thread_locks()
        if thread_locks and str(path) not in thread_locks:
            # Thread holds locks and is trying to acquire a new one
            held_paths = [Path(p) for p in thread_locks.keys()]

            # Check if any other thread is waiting for our locks
            # while holding the lock we want
            with self._global_lock:
                target_key = str(path)
                if target_key in self._held_locks:
                    holder = self._held_locks[target_key]
                    if holder.thread_id != threading.get_ident():
                        # Another thread holds the lock we want
                        # This is a potential deadlock scenario
                        if self._stats:
                            self._stats.record_deadlock()
                        raise DeadlockError(path, held_paths)

    def lock(
        self,
        path: Path,
        mode: LockMode,
        timeout: float | None = None,
        blocking: bool = True,
    ) -> LockHandle:
        """Acquire a lock on the specified path.

        Args:
            path: Path to lock.
            mode: Lock mode (SHARED or EXCLUSIVE).
            timeout: Timeout in seconds. None uses default timeout.
            blocking: If False, fail immediately if lock unavailable.

        Returns:
            LockHandle for the acquired lock.

        Raises:
            LockTimeout: If timeout expires.
            DeadlockError: If potential deadlock detected.
        """
        effective_timeout = timeout if timeout is not None else self._default_timeout
        path_key = str(path)
        thread_locks = self._get_thread_locks()

        # Check for reentry
        if path_key in thread_locks:
            context = thread_locks[path_key]
            # For reentry, must be same or lower lock mode
            if mode == LockMode.EXCLUSIVE and context.handle.mode == LockMode.SHARED:
                raise ValueError("Cannot upgrade shared lock to exclusive")
            context.reentry_count += 1
            return context.handle

        # Check for deadlock
        self._check_deadlock(path)

        # Acquire the lock
        start_time = time.time()
        contended = self._strategy.is_locked(path)

        try:
            handle = self._strategy.acquire(path, mode, effective_timeout, blocking)
        except LockTimeout:
            if self._stats:
                self._stats.record_timeout()
            raise

        wait_time = time.time() - start_time

        # Record the lock
        context = LockContext(handle=handle, wait_time=wait_time)
        thread_locks[path_key] = context

        with self._global_lock:
            self._held_locks[path_key] = context

        # Update statistics
        if self._stats:
            self._stats.record_acquisition(wait_time, contended)
            self._path_stats[path_key].record_acquisition(wait_time, contended)

        return handle

    def unlock(self, handle: LockHandle) -> None:
        """Release a previously acquired lock.

        Args:
            handle: The lock handle to release.

        Raises:
            ValueError: If the lock is not held.
        """
        path_key = str(handle.path)
        thread_locks = self._get_thread_locks()

        if path_key not in thread_locks:
            raise ValueError(f"Lock not held by this thread: {handle.path}")

        context = thread_locks[path_key]

        # Handle reentry
        context.reentry_count -= 1
        if context.reentry_count > 0:
            return

        # Actually release the lock
        hold_time = context.hold_time
        self._strategy.release(handle)

        # Clean up tracking
        del thread_locks[path_key]
        with self._global_lock:
            self._held_locks.pop(path_key, None)

        # Update statistics
        if self._stats:
            self._stats.record_release(hold_time)
            self._path_stats[path_key].record_release(hold_time)

    def try_lock(self, path: Path, mode: LockMode) -> LockHandle | None:
        """Try to acquire a lock without blocking.

        Args:
            path: Path to lock.
            mode: Lock mode.

        Returns:
            LockHandle if acquired, None otherwise.
        """
        try:
            return self.lock(path, mode, timeout=0, blocking=False)
        except (LockTimeout, DeadlockError):
            return None

    def is_locked(self, path: Path) -> bool:
        """Check if a path is currently locked.

        Args:
            path: Path to check.

        Returns:
            True if the path is locked by any thread/process.
        """
        return self._strategy.is_locked(path)

    def is_held_by_current_thread(self, path: Path) -> bool:
        """Check if current thread holds a lock on the path.

        Args:
            path: Path to check.

        Returns:
            True if current thread holds the lock.
        """
        return str(path) in self._get_thread_locks()

    def get_held_locks(self) -> list[LockContext]:
        """Get all locks held by the current thread.

        Returns:
            List of lock contexts.
        """
        return list(self._get_thread_locks().values())

    @contextmanager
    def acquire(
        self,
        path: Path,
        mode: LockMode,
        timeout: float | None = None,
    ) -> Iterator[LockHandle]:
        """Context manager for lock acquisition.

        Args:
            path: Path to lock.
            mode: Lock mode.
            timeout: Timeout in seconds.

        Yields:
            LockHandle for the acquired lock.

        Example:
            >>> with manager.acquire(path, LockMode.EXCLUSIVE) as handle:
            ...     # Safe to modify file
            ...     pass
        """
        handle = self.lock(path, mode, timeout)
        try:
            yield handle
        finally:
            self.unlock(handle)

    @contextmanager
    def read_lock(self, path: Path, timeout: float | None = None) -> Iterator[LockHandle]:
        """Convenience context manager for shared (read) lock.

        Args:
            path: Path to lock.
            timeout: Timeout in seconds.

        Yields:
            LockHandle for the acquired lock.
        """
        with self.acquire(path, LockMode.SHARED, timeout) as handle:
            yield handle

    @contextmanager
    def write_lock(self, path: Path, timeout: float | None = None) -> Iterator[LockHandle]:
        """Convenience context manager for exclusive (write) lock.

        Args:
            path: Path to lock.
            timeout: Timeout in seconds.

        Yields:
            LockHandle for the acquired lock.
        """
        with self.acquire(path, LockMode.EXCLUSIVE, timeout) as handle:
            yield handle

    def release_all(self) -> int:
        """Release all locks held by the current thread.

        Returns:
            Number of locks released.

        Warning:
            This should only be used in cleanup scenarios.
        """
        thread_locks = self._get_thread_locks()
        count = 0

        for path_key in list(thread_locks.keys()):
            context = thread_locks[path_key]
            # Force release regardless of reentry count
            context.reentry_count = 1
            self.unlock(context.handle)
            count += 1

        return count


# Module-level default manager for convenience
_default_manager: FileLockManager | None = None
_manager_lock = threading.Lock()


def get_default_manager() -> FileLockManager:
    """Get the default global lock manager.

    Returns:
        The default FileLockManager instance.
    """
    global _default_manager
    if _default_manager is None:
        with _manager_lock:
            if _default_manager is None:
                _default_manager = FileLockManager()
    return _default_manager


def set_default_manager(manager: FileLockManager) -> None:
    """Set the default global lock manager.

    Args:
        manager: The manager to use as default.
    """
    global _default_manager
    with _manager_lock:
        _default_manager = manager



"""Datetime validators for date format and range checks."""

from truthound.validators.datetime.format import DateFormatValidator
from truthound.validators.datetime.range import (
    DateBetweenValidator,
    FutureDateValidator,
    PastDateValidator,
)
from truthound.validators.datetime.order import DateOrderValidator
from truthound.validators.datetime.timezone import TimezoneValidator
from truthound.validators.datetime.freshness import (
    RecentDataValidator,
    DatePartCoverageValidator,
    GroupedRecentDataValidator,
)
from truthound.validators.datetime.parseable import DateutilParseableValidator

__all__ = [
    "DateFormatValidator",
    "DateBetweenValidator",
    "FutureDateValidator",
    "PastDateValidator",
    "DateOrderValidator",
    "TimezoneValidator",
    "RecentDataValidator",
    "DatePartCoverageValidator",
    "GroupedRecentDataValidator",
    "DateutilParseableValidator",
]



"""Distributed execution framework for large-scale data validation.

This module provides a unified abstraction layer for distributed computing backends
(Spark, Dask, Ray) with:
- Native operations without Polars conversion overhead
- Arrow-based zero-copy data transfer
- Distributed aggregation patterns
- Fault-tolerant execution with checkpointing

Architecture:
    ┌─────────────────────────────────────────────────────────────────────┐
    │                     DistributedExecutionEngine                       │
    │                    (Unified Abstraction Layer)                       │
    ├─────────────────────────────────────────────────────────────────────┤
    │    SparkEngine     │     DaskEngine     │      RayEngine            │
    │  (Native Spark)    │   (Native Dask)    │    (Native Ray)           │
    ├─────────────────────────────────────────────────────────────────────┤
    │                    DistributedBackendProtocol                        │
    │           (Interface for custom backend implementations)             │
    └─────────────────────────────────────────────────────────────────────┘

Example:
    >>> from truthound.execution.distributed import SparkExecutionEngine
    >>>
    >>> # Create Spark-native execution engine
    >>> engine = SparkExecutionEngine.from_dataframe(spark_df)
    >>>
    >>> # Run distributed aggregations
    >>> null_counts = engine.count_nulls_distributed()
    >>> stats = engine.get_stats_distributed("price")
    >>>
    >>> # Convert to Polars only when needed
    >>> lf = engine.to_polars_lazyframe()  # Arrow-based zero-copy
"""

from truthound.execution.distributed.protocols import (
    DistributedBackendProtocol,
    DistributedAggregation,
    PartitionStrategy,
    ExecutionMode,
    DistributedResult,
    PartitionInfo,
    ComputeBackend,
    AggregationScope,
    AggregationSpec,
    BaseAggregator,
    CountAggregator,
    SumAggregator,
    MeanAggregator,
    StdAggregator,
    MinMaxAggregator,
    NullCountAggregator,
    DistinctCountAggregator,
    get_aggregator,
    register_aggregator,
)
from truthound.execution.distributed.base import (
    BaseDistributedEngine,
    DistributedEngineConfig,
    ExecutionMetrics,
)
from truthound.execution.distributed.spark_engine import (
    SparkExecutionEngine,
    SparkEngineConfig,
)
from truthound.execution.distributed.dask_engine import (
    DaskExecutionEngine,
    DaskEngineConfig,
)
from truthound.execution.distributed.ray_engine import (
    RayExecutionEngine,
    RayEngineConfig,
)
from truthound.execution.distributed.aggregations import (
    AggregationPlan,
    AggregationExecutor,
    DistributedAggregator,
    aggregate_distributed,
    create_stats_plan,
    create_null_count_plan,
)
from truthound.execution.distributed.arrow_bridge import (
    ArrowBridge,
    ArrowConversionStrategy,
)
from truthound.execution.distributed.registry import (
    DistributedEngineRegistry,
    get_distributed_engine,
    register_distributed_engine,
    list_distributed_engines,
    list_available_engines,
    get_engine_registry,
)
from truthound.execution.distributed.validator_adapter import (
    DistributedValidatorAdapter,
    AdapterConfig,
    ExecutionStrategy,
    validate_distributed,
    create_distributed_adapter,
)
from truthound.execution.distributed.mixins import (
    StatisticalMixin,
    DataQualityMixin,
    PartitioningMixin,
    IOOperationsMixin,
    ValidationMixin,
    FullFeaturedMixin,
    DataQualityReport,
    ValidationResult,
)

__all__ = [
    # Protocols & Enums
    "DistributedBackendProtocol",
    "DistributedAggregation",
    "PartitionStrategy",
    "ExecutionMode",
    "DistributedResult",
    "PartitionInfo",
    "ComputeBackend",
    "AggregationScope",
    "AggregationSpec",
    # Built-in Aggregators
    "BaseAggregator",
    "CountAggregator",
    "SumAggregator",
    "MeanAggregator",
    "StdAggregator",
    "MinMaxAggregator",
    "NullCountAggregator",
    "DistinctCountAggregator",
    "get_aggregator",
    "register_aggregator",
    # Base classes
    "BaseDistributedEngine",
    "DistributedEngineConfig",
    "ExecutionMetrics",
    # Spark engine
    "SparkExecutionEngine",
    "SparkEngineConfig",
    # Dask engine
    "DaskExecutionEngine",
    "DaskEngineConfig",
    # Ray engine
    "RayExecutionEngine",
    "RayEngineConfig",
    # Aggregations
    "AggregationPlan",
    "AggregationExecutor",
    "DistributedAggregator",
    "aggregate_distributed",
    "create_stats_plan",
    "create_null_count_plan",
    # Arrow bridge
    "ArrowBridge",
    "ArrowConversionStrategy",
    # Registry
    "DistributedEngineRegistry",
    "get_distributed_engine",
    "register_distributed_engine",
    "list_distributed_engines",
    "list_available_engines",
    "get_engine_registry",
    # Validator Adapter
    "DistributedValidatorAdapter",
    "AdapterConfig",
    "ExecutionStrategy",
    "validate_distributed",
    "create_distributed_adapter",
    # Mixins
    "StatisticalMixin",
    "DataQualityMixin",
    "PartitioningMixin",
    "IOOperationsMixin",
    "ValidationMixin",
    "FullFeaturedMixin",
    "DataQualityReport",
    "ValidationResult",
]



"""Rate limiter implementations.

This module provides rate limiting to control the rate of requests
and prevent overload.
"""

from __future__ import annotations

import functools
import logging
import threading
import time
from abc import ABC, abstractmethod
from collections import deque
from contextlib import contextmanager
from typing import Any, Callable, Generator, TypeVar

from truthound.common.resilience.config import RateLimiterConfig
from truthound.common.resilience.protocols import RateLimiterProtocol

logger = logging.getLogger(__name__)

R = TypeVar("R")


class RateLimitExceededError(Exception):
    """Raised when rate limit is exceeded."""

    def __init__(self, name: str, wait_time: float):
        self.limiter_name = name
        self.wait_time = wait_time
        super().__init__(
            f"Rate limit exceeded for '{name}'. Retry after {wait_time:.2f}s"
        )


class RateLimiter(RateLimiterProtocol, ABC):
    """Abstract base class for rate limiter implementations."""

    def __init__(self, name: str, config: RateLimiterConfig | None = None):
        """Initialize rate limiter.

        Args:
            name: Unique name for this rate limiter.
            config: Configuration options.
        """
        self._name = name
        self._config = config or RateLimiterConfig()
        self._total_acquired = 0
        self._total_rejected = 0
        self._lock = threading.Lock()

    @property
    def name(self) -> str:
        """Get rate limiter name."""
        return self._name

    @property
    def config(self) -> RateLimiterConfig:
        """Get configuration."""
        return self._config

    def get_metrics(self) -> dict[str, Any]:
        """Get rate limiter metrics."""
        with self._lock:
            return {
                "name": self._name,
                "rate": self._config.rate,
                "period_seconds": self._config.period_seconds,
                "total_acquired": self._total_acquired,
                "total_rejected": self._total_rejected,
            }

    @contextmanager
    def limit(self, permits: int = 1) -> Generator[None, None, None]:
        """Context manager for rate-limited execution."""
        if not self.acquire(permits):
            wait_time = self.get_wait_time(permits)
            with self._lock:
                self._total_rejected += 1
            raise RateLimitExceededError(self._name, wait_time)

        with self._lock:
            self._total_acquired += permits

        yield

    def __call__(self, func: Callable[..., R]) -> Callable[..., R]:
        """Decorator for rate-limited execution."""
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> R:
            with self.limit():
                return func(*args, **kwargs)
        return wrapper


class TokenBucketRateLimiter(RateLimiter):
    """Token bucket rate limiter.

    Allows bursts up to bucket capacity, then refills at a constant rate.

    Example:
        limiter = TokenBucketRateLimiter(
            "api",
            RateLimiterConfig(rate=100, period_seconds=1.0, burst_size=150)
        )

        @limiter
        def api_call():
            return requests.get(...)
    """

    def __init__(self, name: str, config: RateLimiterConfig | None = None):
        """Initialize token bucket rate limiter."""
        super().__init__(name, config)
        self._tokens = float(self._config.effective_burst_size)
        self._last_refill = time.time()
        self._refill_rate = self._config.rate / self._config.period_seconds

    def acquire(self, permits: int = 1, timeout: float | None = None) -> bool:
        """Acquire permits, blocking if necessary."""
        with self._lock:
            self._refill()

            if self._tokens >= permits:
                self._tokens -= permits
                return True

            if timeout is None or timeout <= 0:
                return False

            # Calculate wait time
            tokens_needed = permits - self._tokens
            wait_time = tokens_needed / self._refill_rate

            if wait_time > timeout:
                return False

        # Wait and retry
        time.sleep(wait_time)

        with self._lock:
            self._refill()
            if self._tokens >= permits:
                self._tokens -= permits
                return True
            return False

    def try_acquire(self, permits: int = 1) -> bool:
        """Try to acquire permits without blocking."""
        with self._lock:
            self._refill()
            if self._tokens >= permits:
                self._tokens -= permits
                return True
            return False

    def get_wait_time(self, permits: int = 1) -> float:
        """Get time to wait before permits are available."""
        with self._lock:
            self._refill()
            if self._tokens >= permits:
                return 0.0
            tokens_needed = permits - self._tokens
            return tokens_needed / self._refill_rate

    def _refill(self) -> None:
        """Refill tokens based on elapsed time."""
        now = time.time()
        elapsed = now - self._last_refill
        tokens_to_add = elapsed * self._refill_rate
        self._tokens = min(
            self._tokens + tokens_to_add,
            float(self._config.effective_burst_size),
        )
        self._last_refill = now

    def reset(self) -> None:
        """Reset rate limiter state."""
        with self._lock:
            self._tokens = float(self._config.effective_burst_size)
            self._last_refill = time.time()
            self._total_acquired = 0
            self._total_rejected = 0


class SlidingWindowRateLimiter(RateLimiter):
    """Sliding window rate limiter.

    Tracks requests in a sliding time window for smooth rate limiting.

    Example:
        limiter = SlidingWindowRateLimiter(
            "api",
            RateLimiterConfig(rate=100, period_seconds=60.0)  # 100 per minute
        )

        @limiter
        def api_call():
            return requests.get(...)
    """

    def __init__(self, name: str, config: RateLimiterConfig | None = None):
        """Initialize sliding window rate limiter."""
        super().__init__(name, config)
        self._timestamps: deque[float] = deque()

    def acquire(self, permits: int = 1, timeout: float | None = None) -> bool:
        """Acquire permits, blocking if necessary."""
        with self._lock:
            self._cleanup_old_entries()

            if len(self._timestamps) + permits <= self._config.rate:
                now = time.time()
                for _ in range(permits):
                    self._timestamps.append(now)
                return True

            if timeout is None or timeout <= 0:
                return False

        # Wait for oldest entries to expire
        wait_time = self.get_wait_time(permits)
        if wait_time > timeout:
            return False

        time.sleep(wait_time)

        with self._lock:
            self._cleanup_old_entries()
            if len(self._timestamps) + permits <= self._config.rate:
                now = time.time()
                for _ in range(permits):
                    self._timestamps.append(now)
                return True
            return False

    def try_acquire(self, permits: int = 1) -> bool:
        """Try to acquire permits without blocking."""
        with self._lock:
            self._cleanup_old_entries()
            if len(self._timestamps) + permits <= self._config.rate:
                now = time.time()
                for _ in range(permits):
                    self._timestamps.append(now)
                return True
            return False

    def get_wait_time(self, permits: int = 1) -> float:
        """Get time to wait before permits are available."""
        with self._lock:
            self._cleanup_old_entries()

            if len(self._timestamps) + permits <= self._config.rate:
                return 0.0

            # How many requests need to expire?
            excess = len(self._timestamps) + permits - self._config.rate

            if excess <= 0:
                return 0.0

            if len(self._timestamps) < excess:
                return self._config.period_seconds

            # Time until oldest required entries expire
            oldest_to_expire = sorted(self._timestamps)[:excess]
            now = time.time()
            expire_times = [
                (ts + self._config.period_seconds) - now
                for ts in oldest_to_expire
            ]
            return max(0.0, max(expire_times))

    def _cleanup_old_entries(self) -> None:
        """Remove entries outside the current window."""
        now = time.time()
        cutoff = now - self._config.period_seconds
        while self._timestamps and self._timestamps[0] < cutoff:
            self._timestamps.popleft()

    def reset(self) -> None:
        """Reset rate limiter state."""
        with self._lock:
            self._timestamps.clear()
            self._total_acquired = 0
            self._total_rejected = 0


class FixedWindowRateLimiter(RateLimiter):
    """Fixed window rate limiter.

    Resets counter at fixed intervals. Simpler but can have edge effects.

    Example:
        limiter = FixedWindowRateLimiter(
            "api",
            RateLimiterConfig(rate=100, period_seconds=60.0)  # 100 per minute
        )

        @limiter
        def api_call():
            return requests.get(...)
    """

    def __init__(self, name: str, config: RateLimiterConfig | None = None):
        """Initialize fixed window rate limiter."""
        super().__init__(name, config)
        self._count = 0
        self._window_start = time.time()

    def acquire(self, permits: int = 1, timeout: float | None = None) -> bool:
        """Acquire permits, blocking if necessary."""
        with self._lock:
            self._check_reset_window()

            if self._count + permits <= self._config.rate:
                self._count += permits
                return True

            if timeout is None or timeout <= 0:
                return False

        # Wait for window reset
        wait_time = self.get_wait_time(permits)
        if wait_time > timeout:
            return False

        time.sleep(wait_time)

        with self._lock:
            self._check_reset_window()
            if self._count + permits <= self._config.rate:
                self._count += permits
                return True
            return False

    def try_acquire(self, permits: int = 1) -> bool:
        """Try to acquire permits without blocking."""
        with self._lock:
            self._check_reset_window()
            if self._count + permits <= self._config.rate:
                self._count += permits
                return True
            return False

    def get_wait_time(self, permits: int = 1) -> float:
        """Get time to wait before permits are available."""
        with self._lock:
            self._check_reset_window()

            if self._count + permits <= self._config.rate:
                return 0.0

            # Wait until window resets
            now = time.time()
            window_end = self._window_start + self._config.period_seconds
            return max(0.0, window_end - now)

    def _check_reset_window(self) -> None:
        """Reset counter if window has passed."""
        now = time.time()
        if now - self._window_start >= self._config.period_seconds:
            self._count = 0
            self._window_start = now

    def reset(self) -> None:
        """Reset rate limiter state."""
        with self._lock:
            self._count = 0
            self._window_start = time.time()
            self._total_acquired = 0
            self._total_rejected = 0


def with_rate_limit(
    rate: int = 100,
    period_seconds: float = 1.0,
    algorithm: str = "token_bucket",
) -> Callable[[Callable[..., R]], Callable[..., R]]:
    """Convenience decorator factory for rate limiting.

    Example:
        @with_rate_limit(rate=10, period_seconds=1.0)
        def limited_operation():
            return api_call()
    """
    config = RateLimiterConfig(
        rate=rate,
        period_seconds=period_seconds,
        algorithm=algorithm,
    )

    if algorithm == "token_bucket":
        limiter = TokenBucketRateLimiter(f"inline-{id(config)}", config)
    elif algorithm == "sliding_window":
        limiter = SlidingWindowRateLimiter(f"inline-{id(config)}", config)
    else:
        limiter = FixedWindowRateLimiter(f"inline-{id(config)}", config)

    def decorator(func: Callable[..., R]) -> Callable[..., R]:
        return limiter(func)

    return decorator



"""Mask command - Mask sensitive data.

This module implements the `truthound mask` command for masking
sensitive data in files.
"""

from __future__ import annotations

from pathlib import Path
from typing import Annotated, Optional

import typer

from truthound.cli_modules.common.errors import error_boundary, require_file
from truthound.cli_modules.common.options import parse_list_callback


@error_boundary
def mask_cmd(
    file: Annotated[
        Path,
        typer.Argument(help="Path to the data file"),
    ],
    output: Annotated[
        Path,
        typer.Option("--output", "-o", help="Output file path"),
    ],
    columns: Annotated[
        Optional[list[str]],
        typer.Option("--columns", "-c", help="Columns to mask (comma-separated)"),
    ] = None,
    strategy: Annotated[
        str,
        typer.Option("--strategy", "-s", help="Masking strategy (redact, hash, fake)"),
    ] = "redact",
) -> None:
    """Mask sensitive data in a file.

    This command creates a copy of the data file with sensitive columns
    masked using the specified strategy.

    Strategies:
        - redact: Replace values with asterisks
        - hash: Replace values with hashed versions
        - fake: Replace values with realistic fake data

    Examples:
        truthound mask data.csv -o masked.csv
        truthound mask data.csv -o masked.csv --columns email,phone
        truthound mask data.csv -o masked.csv --strategy hash
    """
    from truthound.api import mask

    # Validate file exists
    require_file(file)

    # Parse columns if provided
    column_list = parse_list_callback(columns) if columns else None

    try:
        masked_df = mask(str(file), columns=column_list, strategy=strategy)
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(1)

    # Write output based on file extension
    suffix = output.suffix.lower()
    if suffix == ".csv":
        masked_df.write_csv(output)
    elif suffix == ".parquet":
        masked_df.write_parquet(output)
    elif suffix == ".json":
        masked_df.write_json(output)
    else:
        masked_df.write_csv(output)

    typer.echo(f"Masked data written to {output}")



"""Base Classes for Distributed Checkpoint Orchestration.

This module provides abstract base classes that backends can extend
to implement distributed checkpoint execution.
"""

from __future__ import annotations

import asyncio
import logging
import threading
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, Future
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Callable, Generic, Iterator, TypeVar
from queue import Queue, Empty

from truthound.checkpoint.distributed.protocols import (
    BackendCapability,
    ClusterState,
    DistributedBackendProtocol,
    DistributedConfig,
    DistributedError,
    DistributedTask,
    DistributedTaskProtocol,
    DistributedTaskResult,
    TaskCancelledError,
    TaskPriority,
    TaskState,
    TaskSubmissionError,
    TaskTimeoutError,
    WorkerInfo,
    WorkerState,
    WorkerNotAvailableError,
    BackendNotAvailableError,
)

if TYPE_CHECKING:
    from truthound.checkpoint.checkpoint import Checkpoint, CheckpointResult


logger = logging.getLogger(__name__)


T = TypeVar("T")


# =============================================================================
# Metrics Collection
# =============================================================================


@dataclass
class TaskMetrics:
    """Metrics for task execution.

    Attributes:
        tasks_submitted: Total tasks submitted.
        tasks_succeeded: Tasks that succeeded.
        tasks_failed: Tasks that failed.
        tasks_cancelled: Tasks that were cancelled.
        tasks_timed_out: Tasks that timed out.
        total_execution_time_ms: Total execution time in milliseconds.
        total_queue_time_ms: Total time spent in queue.
        retries: Total retry attempts.
    """

    tasks_submitted: int = 0
    tasks_succeeded: int = 0
    tasks_failed: int = 0
    tasks_cancelled: int = 0
    tasks_timed_out: int = 0
    total_execution_time_ms: float = 0.0
    total_queue_time_ms: float = 0.0
    retries: int = 0

    @property
    def tasks_completed(self) -> int:
        """Total completed tasks (success + failed)."""
        return self.tasks_succeeded + self.tasks_failed

    @property
    def success_rate(self) -> float:
        """Success rate (0.0 to 1.0)."""
        if self.tasks_completed == 0:
            return 0.0
        return self.tasks_succeeded / self.tasks_completed

    @property
    def average_execution_time_ms(self) -> float:
        """Average execution time in milliseconds."""
        if self.tasks_completed == 0:
            return 0.0
        return self.total_execution_time_ms / self.tasks_completed

    @property
    def average_queue_time_ms(self) -> float:
        """Average queue time in milliseconds."""
        if self.tasks_completed == 0:
            return 0.0
        return self.total_queue_time_ms / self.tasks_completed

    def record_submission(self) -> None:
        """Record a task submission."""
        self.tasks_submitted += 1

    def record_completion(self, result: DistributedTaskResult) -> None:
        """Record a task completion."""
        if result.state == TaskState.SUCCEEDED:
            self.tasks_succeeded += 1
        elif result.state == TaskState.FAILED:
            self.tasks_failed += 1
        elif result.state == TaskState.CANCELLED:
            self.tasks_cancelled += 1
        elif result.state == TaskState.TIMEOUT:
            self.tasks_timed_out += 1

        if result.duration_ms:
            self.total_execution_time_ms += result.duration_ms
        if result.queue_time_ms:
            self.total_queue_time_ms += result.queue_time_ms
        self.retries += result.retries

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            "tasks_submitted": self.tasks_submitted,
            "tasks_succeeded": self.tasks_succeeded,
            "tasks_failed": self.tasks_failed,
            "tasks_cancelled": self.tasks_cancelled,
            "tasks_timed_out": self.tasks_timed_out,
            "tasks_completed": self.tasks_completed,
            "success_rate": self.success_rate,
            "average_execution_time_ms": self.average_execution_time_ms,
            "average_queue_time_ms": self.average_queue_time_ms,
            "total_retries": self.retries,
        }

    def to_prometheus(self) -> str:
        """Export as Prometheus metrics."""
        lines = [
            f"truthound_distributed_tasks_submitted_total {self.tasks_submitted}",
            f"truthound_distributed_tasks_succeeded_total {self.tasks_succeeded}",
            f"truthound_distributed_tasks_failed_total {self.tasks_failed}",
            f"truthound_distributed_tasks_cancelled_total {self.tasks_cancelled}",
            f"truthound_distributed_tasks_timed_out_total {self.tasks_timed_out}",
            f"truthound_distributed_task_execution_time_ms_total {self.total_execution_time_ms}",
            f"truthound_distributed_task_queue_time_ms_total {self.total_queue_time_ms}",
            f"truthound_distributed_task_retries_total {self.retries}",
        ]
        return "\n".join(lines)


# =============================================================================
# Base Backend Implementation
# =============================================================================


class BaseDistributedBackend(ABC):
    """Abstract base class for distributed backends.

    This provides common functionality that all backends can use,
    including metrics collection, retry logic, and task tracking.

    Subclasses must implement:
        - _do_submit(): Actual task submission logic
        - _do_connect(): Backend connection logic
        - _do_disconnect(): Backend disconnection logic
        - _do_get_cluster_state(): Get cluster state
        - _do_get_workers(): Get worker list
    """

    def __init__(
        self,
        config: DistributedConfig | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the backend.

        Args:
            config: Distributed configuration.
            **kwargs: Additional backend-specific options.
        """
        self._config = config or DistributedConfig()

        # Apply kwargs to backend_options
        for key, value in kwargs.items():
            self._config.backend_options[key] = value

        self._connected = False
        self._lock = threading.RLock()
        self._tasks: dict[str, DistributedTask[Any]] = {}
        self._metrics = TaskMetrics()

        # Callbacks
        self._on_task_complete: list[Callable[[DistributedTaskResult], None]] = []
        self._on_task_error: list[Callable[[str, Exception], None]] = []

    @property
    @abstractmethod
    def name(self) -> str:
        """Get backend name."""
        ...

    @property
    @abstractmethod
    def capabilities(self) -> BackendCapability:
        """Get backend capabilities."""
        ...

    @property
    def is_connected(self) -> bool:
        """Check if backend is connected."""
        return self._connected

    @property
    def config(self) -> DistributedConfig:
        """Get backend configuration."""
        return self._config

    @property
    def metrics(self) -> TaskMetrics:
        """Get task metrics."""
        return self._metrics

    # -------------------------------------------------------------------------
    # Connection Management
    # -------------------------------------------------------------------------

    def connect(self, **kwargs: Any) -> None:
        """Connect to the backend."""
        with self._lock:
            if self._connected:
                return

            try:
                self._do_connect(**kwargs)
                self._connected = True
                logger.info(f"Connected to {self.name} backend")
            except Exception as e:
                raise BackendNotAvailableError(
                    self.name,
                    reason=str(e),
                )

    def disconnect(self) -> None:
        """Disconnect from the backend."""
        with self._lock:
            if not self._connected:
                return

            try:
                self._do_disconnect()
            finally:
                self._connected = False
                logger.info(f"Disconnected from {self.name} backend")

    @abstractmethod
    def _do_connect(self, **kwargs: Any) -> None:
        """Perform actual connection (implemented by subclass)."""
        ...

    @abstractmethod
    def _do_disconnect(self) -> None:
        """Perform actual disconnection (implemented by subclass)."""
        ...

    @contextmanager
    def connection(self, **kwargs: Any) -> Iterator[None]:
        """Context manager for connection lifecycle."""
        self.connect(**kwargs)
        try:
            yield
        finally:
            self.disconnect()

    # -------------------------------------------------------------------------
    # Task Submission
    # -------------------------------------------------------------------------

    def submit(
        self,
        checkpoint: "Checkpoint",
        priority: TaskPriority = TaskPriority.NORMAL,
        timeout: float | None = None,
        context: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> DistributedTaskProtocol["CheckpointResult"]:
        """Submit a checkpoint for execution."""
        if not self._connected:
            raise TaskSubmissionError(
                "Backend not connected",
                checkpoint_name=checkpoint.name,
                reason="Call connect() first",
            )

        # Create task
        task: DistributedTask["CheckpointResult"] = DistributedTask.create(
            checkpoint=checkpoint,
            backend=self,
        )
        task._metadata["priority"] = priority.value
        task._metadata["context"] = context or {}

        timeout = timeout or self._config.task_timeout_seconds

        with self._lock:
            self._tasks[task.task_id] = task
            self._metrics.record_submission()

        try:
            # Submit to backend
            task._set_state(TaskState.SUBMITTED)
            self._do_submit(task, checkpoint, priority, timeout, context, **kwargs)
            task._set_state(TaskState.QUEUED)
            logger.debug(f"Task {task.task_id} submitted for checkpoint {checkpoint.name}")
            return task

        except Exception as e:
            task._set_error(str(e), e)
            raise TaskSubmissionError(
                f"Failed to submit task: {e}",
                checkpoint_name=checkpoint.name,
                reason=str(e),
            )

    def submit_batch(
        self,
        checkpoints: list["Checkpoint"],
        priority: TaskPriority = TaskPriority.NORMAL,
        timeout: float | None = None,
        context: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[DistributedTaskProtocol["CheckpointResult"]]:
        """Submit multiple checkpoints for execution."""
        # Check capability
        if not (self.capabilities & BackendCapability.BATCH_SUBMIT):
            # Fall back to sequential submission
            return [
                self.submit(cp, priority, timeout, context, **kwargs)
                for cp in checkpoints
            ]

        tasks = []
        for checkpoint in checkpoints:
            try:
                task = self.submit(checkpoint, priority, timeout, context, **kwargs)
                tasks.append(task)
            except TaskSubmissionError as e:
                logger.error(f"Failed to submit {checkpoint.name}: {e}")
                # Continue with other checkpoints

        return tasks

    @abstractmethod
    def _do_submit(
        self,
        task: DistributedTask["CheckpointResult"],
        checkpoint: "Checkpoint",
        priority: TaskPriority,
        timeout: float,
        context: dict[str, Any] | None,
        **kwargs: Any,
    ) -> None:
        """Perform actual task submission (implemented by subclass)."""
        ...

    # -------------------------------------------------------------------------
    # Task Management
    # -------------------------------------------------------------------------

    def get_task(self, task_id: str) -> DistributedTaskProtocol["CheckpointResult"] | None:
        """Get a task by ID."""
        with self._lock:
            return self._tasks.get(task_id)

    def cancel_task(self, task_id: str, terminate: bool = False) -> bool:
        """Cancel a task."""
        with self._lock:
            task = self._tasks.get(task_id)
            if task is None:
                return False
            return task.cancel(terminate=terminate)

    def _complete_task(
        self,
        task_id: str,
        result: "CheckpointResult | None" = None,
        error: str | None = None,
        exception: Exception | None = None,
    ) -> None:
        """Mark a task as complete (internal use)."""
        with self._lock:
            task = self._tasks.get(task_id)
            if task is None:
                return

            if error:
                task._set_error(error, exception)
            elif result:
                task._set_result(result)

            # Record metrics
            self._metrics.record_completion(task.to_result())

            # Invoke callbacks
            task_result = task.to_result()
            for callback in self._on_task_complete:
                try:
                    callback(task_result)
                except Exception as e:
                    logger.error(f"Callback error: {e}")

    # -------------------------------------------------------------------------
    # Cluster State
    # -------------------------------------------------------------------------

    def get_cluster_state(self) -> ClusterState:
        """Get current cluster state."""
        if not self._connected:
            return ClusterState(
                backend_name=self.name,
                is_healthy=False,
            )
        return self._do_get_cluster_state()

    @abstractmethod
    def _do_get_cluster_state(self) -> ClusterState:
        """Get cluster state (implemented by subclass)."""
        ...

    def get_workers(self) -> list[WorkerInfo]:
        """Get list of workers."""
        if not self._connected:
            return []
        return self._do_get_workers()

    @abstractmethod
    def _do_get_workers(self) -> list[WorkerInfo]:
        """Get workers (implemented by subclass)."""
        ...

    def scale_workers(self, count: int) -> bool:
        """Scale the number of workers."""
        if not (self.capabilities & BackendCapability.WORKER_SCALING):
            return False
        return self._do_scale_workers(count)

    def _do_scale_workers(self, count: int) -> bool:
        """Scale workers (override if supported)."""
        return False

    def health_check(self) -> bool:
        """Perform health check."""
        if not self._connected:
            return False
        try:
            state = self.get_cluster_state()
            return state.is_healthy
        except Exception:
            return False

    # -------------------------------------------------------------------------
    # Callbacks
    # -------------------------------------------------------------------------

    def on_task_complete(
        self,
        callback: Callable[[DistributedTaskResult], None],
    ) -> None:
        """Register a task completion callback."""
        self._on_task_complete.append(callback)

    def on_task_error(
        self,
        callback: Callable[[str, Exception], None],
    ) -> None:
        """Register a task error callback."""
        self._on_task_error.append(callback)

    # -------------------------------------------------------------------------
    # Utilities
    # -------------------------------------------------------------------------

    def _calculate_retry_delay(self, attempt: int) -> float:
        """Calculate retry delay with exponential backoff."""
        delay = self._config.retry_delay_seconds * (
            self._config.retry_backoff_multiplier ** attempt
        )
        return min(delay, self._config.retry_max_delay_seconds)


# =============================================================================
# Base Orchestrator Implementation
# =============================================================================


class BaseDistributedOrchestrator:
    """Base class for distributed checkpoint orchestrators.

    The orchestrator provides a high-level API for submitting and
    managing distributed checkpoint execution across multiple backends.
    """

    def __init__(
        self,
        backend: BaseDistributedBackend,
        config: DistributedConfig | None = None,
    ) -> None:
        """Initialize the orchestrator.

        Args:
            backend: The distributed backend to use.
            config: Configuration options.
        """
        self._backend = backend
        self._config = config or backend.config
        self._lock = threading.RLock()
        self._executor: ThreadPoolExecutor | None = None

    @property
    def backend(self) -> BaseDistributedBackend:
        """Get the backend."""
        return self._backend

    @property
    def is_connected(self) -> bool:
        """Check if connected to backend."""
        return self._backend.is_connected

    @property
    def metrics(self) -> TaskMetrics:
        """Get execution metrics."""
        return self._backend.metrics

    # -------------------------------------------------------------------------
    # Connection
    # -------------------------------------------------------------------------

    def connect(self, **kwargs: Any) -> "BaseDistributedOrchestrator":
        """Connect to the backend."""
        self._backend.connect(**kwargs)
        return self

    def disconnect(self) -> None:
        """Disconnect from the backend."""
        self._backend.disconnect()
        if self._executor:
            self._executor.shutdown(wait=False)
            self._executor = None

    def __enter__(self) -> "BaseDistributedOrchestrator":
        self.connect()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.disconnect()

    # -------------------------------------------------------------------------
    # Task Submission
    # -------------------------------------------------------------------------

    def submit(
        self,
        checkpoint: "Checkpoint",
        priority: int | TaskPriority = TaskPriority.NORMAL,
        timeout: float | None = None,
        context: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> DistributedTaskProtocol["CheckpointResult"]:
        """Submit a checkpoint for distributed execution.

        Args:
            checkpoint: Checkpoint to execute.
            priority: Task priority (0-10 or TaskPriority).
            timeout: Task timeout in seconds.
            context: Additional context for the run.
            **kwargs: Backend-specific options.

        Returns:
            Task handle for tracking execution.
        """
        if isinstance(priority, int):
            priority = TaskPriority.from_int(priority)

        return self._backend.submit(
            checkpoint=checkpoint,
            priority=priority,
            timeout=timeout,
            context=context,
            **kwargs,
        )

    def submit_batch(
        self,
        checkpoints: list["Checkpoint"],
        priority: int | TaskPriority = TaskPriority.NORMAL,
        timeout: float | None = None,
        context: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[DistributedTaskProtocol["CheckpointResult"]]:
        """Submit multiple checkpoints for execution.

        Args:
            checkpoints: List of checkpoints to execute.
            priority: Task priority.
            timeout: Task timeout in seconds.
            context: Additional context for all runs.
            **kwargs: Backend-specific options.

        Returns:
            List of task handles.
        """
        if isinstance(priority, int):
            priority = TaskPriority.from_int(priority)

        return self._backend.submit_batch(
            checkpoints=checkpoints,
            priority=priority,
            timeout=timeout,
            context=context,
            **kwargs,
        )

    # -------------------------------------------------------------------------
    # Result Collection
    # -------------------------------------------------------------------------

    def gather(
        self,
        tasks: list[DistributedTaskProtocol["CheckpointResult"]],
        timeout: float | None = None,
        return_exceptions: bool = False,
    ) -> list["CheckpointResult | Exception"]:
        """Wait for all tasks to complete and return results.

        Args:
            tasks: List of tasks to wait for.
            timeout: Maximum time to wait.
            return_exceptions: If True, return exceptions instead of raising.

        Returns:
            List of results (or exceptions if return_exceptions=True).
        """
        results: list["CheckpointResult | Exception"] = []
        deadline = time.time() + timeout if timeout else None

        for task in tasks:
            remaining = None
            if deadline:
                remaining = max(0, deadline - time.time())
                if remaining == 0:
                    if return_exceptions:
                        results.append(
                            TaskTimeoutError("Gather timeout", task.task_id, timeout or 0)
                        )
                        continue
                    raise TaskTimeoutError("Gather timeout", task.task_id, timeout or 0)

            try:
                result = task.result(timeout=remaining)
                results.append(result)
            except Exception as e:
                if return_exceptions:
                    results.append(e)
                else:
                    raise

        return results

    async def gather_async(
        self,
        tasks: list[DistributedTaskProtocol["CheckpointResult"]],
        timeout: float | None = None,
        return_exceptions: bool = False,
    ) -> list["CheckpointResult | Exception"]:
        """Async version of gather()."""
        results: list["CheckpointResult | Exception"] = []

        async def get_result(
            task: DistributedTaskProtocol["CheckpointResult"],
        ) -> "CheckpointResult | Exception":
            try:
                return await task.result_async(timeout=timeout)
            except Exception as e:
                if return_exceptions:
                    return e
                raise

        gathered = await asyncio.gather(
            *[get_result(t) for t in tasks],
            return_exceptions=return_exceptions,
        )
        return list(gathered)

    def as_completed(
        self,
        tasks: list[DistributedTaskProtocol["CheckpointResult"]],
        timeout: float | None = None,
    ) -> Iterator[DistributedTaskProtocol["CheckpointResult"]]:
        """Yield tasks as they complete.

        Args:
            tasks: List of tasks to monitor.
            timeout: Maximum time to wait for each task.

        Yields:
            Tasks as they complete.
        """
        pending = set(tasks)
        deadline = time.time() + timeout if timeout else None

        while pending:
            completed = None
            for task in pending:
                if task.is_ready():
                    completed = task
                    break

            if completed:
                pending.remove(completed)
                yield completed
            else:
                if deadline and time.time() > deadline:
                    return
                time.sleep(0.05)

    # -------------------------------------------------------------------------
    # Cluster Management
    # -------------------------------------------------------------------------

    def get_cluster_state(self) -> ClusterState:
        """Get current cluster state."""
        return self._backend.get_cluster_state()

    def get_workers(self) -> list[WorkerInfo]:
        """Get list of workers."""
        return self._backend.get_workers()

    def scale_workers(self, count: int) -> bool:
        """Scale the number of workers."""
        return self._backend.scale_workers(count)

    def health_check(self) -> bool:
        """Check cluster health."""
        return self._backend.health_check()

    def wait_for_workers(
        self,
        min_workers: int = 1,
        timeout: float = 60.0,
    ) -> bool:
        """Wait for minimum number of workers to be available.

        Args:
            min_workers: Minimum number of workers required.
            timeout: Maximum time to wait.

        Returns:
            True if workers are available, False if timeout.
        """
        deadline = time.time() + timeout

        while time.time() < deadline:
            workers = self.get_workers()
            available = len([w for w in workers if w.is_available])
            if available >= min_workers:
                return True
            time.sleep(1.0)

        return False



"""Tiered storage system for validation stores.

This module provides a tiered storage layer that automatically moves data
between different storage tiers based on access patterns, age, and policies.

Example:
    >>> from truthound.stores import get_store
    >>> from truthound.stores.tiering import (
    ...     TieredStore,
    ...     TieringConfig,
    ...     StorageTier,
    ...     TierPolicy,
    ... )
    >>>
    >>> # Define storage tiers
    >>> tiers = [
    ...     StorageTier(
    ...         name="hot",
    ...         store=get_store("filesystem", base_path=".truthound/hot"),
    ...         priority=1,  # Highest priority, fastest access
    ...     ),
    ...     StorageTier(
    ...         name="warm",
    ...         store=get_store("s3", bucket="my-bucket", storage_class="STANDARD_IA"),
    ...         priority=2,
    ...     ),
    ...     StorageTier(
    ...         name="cold",
    ...         store=get_store("s3", bucket="my-bucket", storage_class="GLACIER"),
    ...         priority=3,  # Lowest priority, cheapest
    ...     ),
    ... ]
    >>>
    >>> # Define tiering policies
    >>> policies = [
    ...     TierPolicy(
    ...         from_tier="hot",
    ...         to_tier="warm",
    ...         after_days=7,
    ...     ),
    ...     TierPolicy(
    ...         from_tier="warm",
    ...         to_tier="cold",
    ...         after_days=30,
    ...     ),
    ... ]
    >>>
    >>> # Create tiered store
    >>> store = TieredStore(tiers, TieringConfig(policies=policies))
    >>>
    >>> # Use like any other store
    >>> store.save(result)  # Saved to hot tier
    >>> store.get(run_id)  # Fetched from appropriate tier
"""

from truthound.stores.tiering.base import (
    StorageTier,
    TierInfo,
    TierType,
    TieringConfig,
    TierPolicy,
    TieringResult,
    TieringError,
    TierNotFoundError,
    TierMigrationError,
    TierAccessError,
    MigrationDirection,
    TierMetadataStore,
    InMemoryTierMetadataStore,
)
from truthound.stores.tiering.store import TieredStore
from truthound.stores.tiering.policies import (
    AgeBasedTierPolicy,
    AccessBasedTierPolicy,
    SizeBasedTierPolicy,
    ScheduledTierPolicy,
    CompositeTierPolicy,
)
from truthound.stores.tiering.manager import TieringManager

__all__ = [
    # Base types
    "StorageTier",
    "TierInfo",
    "TierType",
    "TieringConfig",
    "TierPolicy",
    "TieringResult",
    "MigrationDirection",
    # Errors
    "TieringError",
    "TierNotFoundError",
    "TierMigrationError",
    "TierAccessError",
    # Metadata stores
    "TierMetadataStore",
    "InMemoryTierMetadataStore",
    # Policies
    "AgeBasedTierPolicy",
    "AccessBasedTierPolicy",
    "SizeBasedTierPolicy",
    "ScheduledTierPolicy",
    "CompositeTierPolicy",
    # Main store
    "TieredStore",
    # Manager
    "TieringManager",
]



"""Event-based triggers.

These triggers execute checkpoints based on events like file changes,
webhooks, or custom events.
"""

from __future__ import annotations

import fnmatch
import hashlib
import os
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from truthound.checkpoint.triggers.base import (
    BaseTrigger,
    TriggerConfig,
    TriggerResult,
)

if TYPE_CHECKING:
    pass


@dataclass
class EventConfig(TriggerConfig):
    """Configuration for event trigger.

    Attributes:
        event_type: Type of event to listen for.
        event_filter: Filter for matching events.
        debounce_seconds: Minimum time between trigger fires.
        batch_events: Batch multiple events into single run.
        batch_window_seconds: Time window for batching events.
    """

    event_type: str = "custom"
    event_filter: dict[str, Any] = field(default_factory=dict)
    debounce_seconds: float = 1.0
    batch_events: bool = False
    batch_window_seconds: float = 5.0


class EventTrigger(BaseTrigger[EventConfig]):
    """Trigger based on custom events.

    Executes checkpoints when specific events occur. Events can be
    pushed programmatically or received from external sources.

    Example:
        >>> trigger = EventTrigger(
        ...     event_type="data_updated",
        ...     event_filter={"source": "production"},
        ... )
        >>>
        >>> # Fire the trigger
        >>> trigger.fire_event({"source": "production", "table": "users"})
    """

    trigger_type = "event"

    def __init__(self, config: EventConfig | None = None, **kwargs: Any) -> None:
        super().__init__(config, **kwargs)
        self._pending_events: list[dict[str, Any]] = []
        self._event_callback: Callable[[dict[str, Any]], None] | None = None
        self._last_event_time: datetime | None = None

    @classmethod
    def _default_config(cls) -> EventConfig:
        return EventConfig()

    def fire_event(self, event_data: dict[str, Any]) -> bool:
        """Fire an event to trigger the checkpoint.

        Args:
            event_data: Data associated with the event.

        Returns:
            True if event was accepted, False if filtered out.
        """
        # Check filter
        filter_config = self._config.event_filter
        if filter_config:
            for key, expected in filter_config.items():
                if event_data.get(key) != expected:
                    return False

        # Check debounce
        now = datetime.now()
        if self._last_event_time:
            elapsed = (now - self._last_event_time).total_seconds()
            if elapsed < self._config.debounce_seconds:
                if self._config.batch_events:
                    self._pending_events.append(event_data)
                return False

        self._last_event_time = now
        self._pending_events.append(event_data)

        if self._event_callback:
            self._event_callback(event_data)

        return True

    def should_trigger(self) -> TriggerResult:
        """Check if there are pending events."""
        if not self._pending_events:
            return TriggerResult(
                should_run=False,
                reason="No pending events",
            )

        # Check batch window
        if self._config.batch_events:
            if self._last_event_time:
                elapsed = (datetime.now() - self._last_event_time).total_seconds()
                if elapsed < self._config.batch_window_seconds:
                    return TriggerResult(
                        should_run=False,
                        reason=f"Batching events, {elapsed:.1f}s elapsed",
                        context={"pending_count": len(self._pending_events)},
                    )

        # Consume events
        events = self._pending_events.copy()
        self._pending_events.clear()

        return TriggerResult(
            should_run=True,
            reason=f"Received {len(events)} event(s)",
            context={"events": events},
        )

    def on_event(self, callback: Callable[[dict[str, Any]], None]) -> None:
        """Register callback for event notifications.

        Args:
            callback: Function to call when events arrive.
        """
        self._event_callback = callback

    def clear_pending(self) -> int:
        """Clear pending events.

        Returns:
            Number of cleared events.
        """
        count = len(self._pending_events)
        self._pending_events.clear()
        return count


@dataclass
class FileWatchConfig(TriggerConfig):
    """Configuration for file watch trigger.

    Attributes:
        paths: Paths to watch (files or directories).
        patterns: Glob patterns to match (e.g., "*.csv").
        recursive: Watch directories recursively.
        events: File events to watch ("created", "modified", "deleted").
        ignore_patterns: Patterns to ignore.
        hash_check: Only trigger if file content changed (hash check).
        poll_interval_seconds: Polling interval for checking changes.
    """

    paths: list[str] = field(default_factory=list)
    patterns: list[str] = field(default_factory=lambda: ["*"])
    recursive: bool = True
    events: list[str] = field(default_factory=lambda: ["modified", "created"])
    ignore_patterns: list[str] = field(default_factory=lambda: [".*", "__pycache__", "*.pyc"])
    hash_check: bool = True
    poll_interval_seconds: float = 5.0


class FileWatchTrigger(BaseTrigger[FileWatchConfig]):
    """Trigger based on file system changes.

    Executes checkpoints when files matching specified patterns
    are created, modified, or deleted.

    Example:
        >>> trigger = FileWatchTrigger(
        ...     paths=["./data"],
        ...     patterns=["*.csv", "*.parquet"],
        ...     events=["modified", "created"],
        ... )
    """

    trigger_type = "file_watch"

    def __init__(self, config: FileWatchConfig | None = None, **kwargs: Any) -> None:
        super().__init__(config, **kwargs)
        self._file_hashes: dict[str, str] = {}
        self._file_mtimes: dict[str, float] = {}
        self._last_scan: datetime | None = None
        self._pending_changes: list[dict[str, Any]] = []

    @classmethod
    def _default_config(cls) -> FileWatchConfig:
        return FileWatchConfig()

    def _get_file_hash(self, filepath: Path) -> str:
        """Calculate MD5 hash of file content."""
        hasher = hashlib.md5()
        try:
            with open(filepath, "rb") as f:
                for chunk in iter(lambda: f.read(8192), b""):
                    hasher.update(chunk)
            return hasher.hexdigest()
        except (IOError, OSError):
            return ""

    def _matches_pattern(self, filepath: Path) -> bool:
        """Check if file matches any include pattern and no ignore pattern."""
        filename = filepath.name

        # Check ignore patterns
        for pattern in self._config.ignore_patterns:
            if fnmatch.fnmatch(filename, pattern):
                return False
            if fnmatch.fnmatch(str(filepath), pattern):
                return False

        # Check include patterns
        for pattern in self._config.patterns:
            if fnmatch.fnmatch(filename, pattern):
                return True
            if fnmatch.fnmatch(str(filepath), pattern):
                return True

        return False

    def _scan_files(self) -> list[Path]:
        """Scan configured paths for matching files."""
        files: list[Path] = []

        for path_str in self._config.paths:
            path = Path(path_str).expanduser().resolve()

            if path.is_file():
                if self._matches_pattern(path):
                    files.append(path)
            elif path.is_dir():
                if self._config.recursive:
                    for filepath in path.rglob("*"):
                        if filepath.is_file() and self._matches_pattern(filepath):
                            files.append(filepath)
                else:
                    for filepath in path.glob("*"):
                        if filepath.is_file() and self._matches_pattern(filepath):
                            files.append(filepath)

        return files

    def _detect_changes(self) -> list[dict[str, Any]]:
        """Detect file changes since last scan."""
        changes: list[dict[str, Any]] = []
        current_files = self._scan_files()
        current_file_set = set(str(f) for f in current_files)
        previous_file_set = set(self._file_mtimes.keys())

        # Check for new and modified files
        for filepath in current_files:
            filepath_str = str(filepath)
            try:
                mtime = filepath.stat().st_mtime
            except (IOError, OSError):
                continue

            if filepath_str not in self._file_mtimes:
                # New file
                if "created" in self._config.events:
                    file_hash = self._get_file_hash(filepath) if self._config.hash_check else ""
                    self._file_hashes[filepath_str] = file_hash
                    self._file_mtimes[filepath_str] = mtime
                    changes.append({
                        "event": "created",
                        "path": filepath_str,
                        "mtime": mtime,
                    })
            elif mtime != self._file_mtimes[filepath_str]:
                # Potentially modified
                if "modified" in self._config.events:
                    if self._config.hash_check:
                        new_hash = self._get_file_hash(filepath)
                        old_hash = self._file_hashes.get(filepath_str, "")
                        if new_hash != old_hash:
                            self._file_hashes[filepath_str] = new_hash
                            self._file_mtimes[filepath_str] = mtime
                            changes.append({
                                "event": "modified",
                                "path": filepath_str,
                                "mtime": mtime,
                            })
                        else:
                            # mtime changed but content didn't
                            self._file_mtimes[filepath_str] = mtime
                    else:
                        self._file_mtimes[filepath_str] = mtime
                        changes.append({
                            "event": "modified",
                            "path": filepath_str,
                            "mtime": mtime,
                        })

        # Check for deleted files
        if "deleted" in self._config.events:
            for filepath_str in previous_file_set - current_file_set:
                del self._file_mtimes[filepath_str]
                if filepath_str in self._file_hashes:
                    del self._file_hashes[filepath_str]
                changes.append({
                    "event": "deleted",
                    "path": filepath_str,
                })

        return changes

    def should_trigger(self) -> TriggerResult:
        """Check for file system changes."""
        now = datetime.now()

        # Check poll interval
        if self._last_scan:
            elapsed = (now - self._last_scan).total_seconds()
            if elapsed < self._config.poll_interval_seconds:
                return TriggerResult(
                    should_run=False,
                    reason=f"Waiting for poll interval ({elapsed:.1f}s elapsed)",
                )

        self._last_scan = now

        # Detect changes
        changes = self._detect_changes()

        if not changes:
            return TriggerResult(
                should_run=False,
                reason="No file changes detected",
            )

        return TriggerResult(
            should_run=True,
            reason=f"Detected {len(changes)} file change(s)",
            context={
                "changes": changes,
                "changed_files": [c["path"] for c in changes],
            },
        )

    def _on_start(self) -> None:
        """Initialize file tracking on start."""
        # Initial scan to establish baseline
        for filepath in self._scan_files():
            filepath_str = str(filepath)
            try:
                self._file_mtimes[filepath_str] = filepath.stat().st_mtime
                if self._config.hash_check:
                    self._file_hashes[filepath_str] = self._get_file_hash(filepath)
            except (IOError, OSError):
                pass

        self._last_scan = datetime.now()

    def validate_config(self) -> list[str]:
        """Validate configuration."""
        errors = []

        if not self._config.paths:
            errors.append("At least one path is required")

        for path_str in self._config.paths:
            path = Path(path_str).expanduser()
            if not path.exists():
                errors.append(f"Path does not exist: {path_str}")

        for event in self._config.events:
            if event not in ("created", "modified", "deleted"):
                errors.append(f"Invalid event type: {event}")

        return errors



"""Middleware and decorators for RBAC.

This module provides middleware for web frameworks and decorators
for permission-protected function execution.
"""

from __future__ import annotations

import functools
import time
from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, TypeVar

from truthound.rbac.core import (
    AccessContext,
    AccessDecision,
    Permission,
    PermissionAction,
    PermissionDeniedError,
    Principal,
    PrincipalStore,
    ResourceType,
    RoleStore,
    SecurityContext,
)
from truthound.rbac.policy import PolicyEngine


F = TypeVar("F", bound=Callable[..., Any])
AsyncF = TypeVar("AsyncF", bound=Callable[..., Awaitable[Any]])


# =============================================================================
# Configuration
# =============================================================================


@dataclass
class RBACMiddlewareConfig:
    """Configuration for RBAC middleware."""

    # Principal resolution
    principal_header: str = "X-Principal-ID"
    principal_from_jwt: bool = True
    jwt_claim: str = "sub"

    # Default behavior
    deny_anonymous: bool = True
    allow_public_paths: list[str] = field(
        default_factory=lambda: ["/health", "/ready", "/metrics"]
    )

    # Error handling
    raise_on_denied: bool = True
    denied_status_code: int = 403

    # Logging
    log_access_decisions: bool = True

    # Audit
    audit_decisions: bool = True


# =============================================================================
# RBAC Middleware Base
# =============================================================================


class RBACMiddleware:
    """Base middleware for RBAC enforcement.

    Example:
        >>> middleware = RBACMiddleware(
        ...     engine=policy_engine,
        ...     principal_store=principal_store,
        ... )
    """

    def __init__(
        self,
        engine: PolicyEngine,
        principal_store: PrincipalStore,
        config: RBACMiddlewareConfig | None = None,
    ) -> None:
        self._engine = engine
        self._principal_store = principal_store
        self._config = config or RBACMiddlewareConfig()

    def resolve_principal(self, context: dict[str, Any]) -> Principal | None:
        """Resolve principal from request context."""
        headers = context.get("headers", {})

        # Try header-based resolution
        principal_id = headers.get(self._config.principal_header)
        if principal_id:
            return self._principal_store.get(principal_id)

        # Try JWT-based resolution
        if self._config.principal_from_jwt:
            auth_header = headers.get("Authorization", "")
            if auth_header.startswith("Bearer "):
                principal_id = self._extract_from_jwt(auth_header[7:])
                if principal_id:
                    return self._principal_store.get(principal_id)

        return None

    def _extract_from_jwt(self, token: str) -> str | None:
        """Extract principal ID from JWT token."""
        import base64
        import json

        try:
            parts = token.split(".")
            if len(parts) != 3:
                return None

            payload = parts[1]
            # Add padding
            padding = 4 - len(payload) % 4
            if padding != 4:
                payload += "=" * padding

            decoded = base64.urlsafe_b64decode(payload)
            claims = json.loads(decoded)
            return claims.get(self._config.jwt_claim)
        except Exception:
            return None

    def is_public_path(self, path: str) -> bool:
        """Check if the path is public (no auth required)."""
        for public_path in self._config.allow_public_paths:
            if path.startswith(public_path):
                return True
        return False

    def check_access(
        self,
        principal: Principal | None,
        resource: str,
        action: str,
        resource_attributes: dict[str, Any] | None = None,
    ) -> AccessDecision:
        """Check access for the principal."""
        if principal is None:
            if self._config.deny_anonymous:
                return AccessDecision.deny("Anonymous access denied")
            principal = Principal.anonymous()

        return self._engine.check(
            principal=principal,
            resource=resource,
            action=action,
            resource_attributes=resource_attributes,
        )


# =============================================================================
# ASGI Middleware
# =============================================================================


class ASGIRBACMiddleware:
    """ASGI middleware for RBAC enforcement.

    Example:
        >>> from fastapi import FastAPI
        >>> app = FastAPI()
        >>> app.add_middleware(
        ...     ASGIRBACMiddleware,
        ...     engine=policy_engine,
        ...     principal_store=principal_store,
        ... )
    """

    def __init__(
        self,
        app: Any,
        engine: PolicyEngine,
        principal_store: PrincipalStore,
        config: RBACMiddlewareConfig | None = None,
        resource_mapper: Callable[[str, str], tuple[str, str]] | None = None,
    ) -> None:
        self.app = app
        self._middleware = RBACMiddleware(
            engine=engine,
            principal_store=principal_store,
            config=config,
        )
        self._config = config or RBACMiddlewareConfig()
        self._resource_mapper = resource_mapper or self._default_resource_mapper

    def _default_resource_mapper(self, path: str, method: str) -> tuple[str, str]:
        """Default resource/action mapper from HTTP path and method."""
        # Map HTTP method to action
        method_to_action = {
            "GET": "read",
            "HEAD": "read",
            "POST": "create",
            "PUT": "update",
            "PATCH": "update",
            "DELETE": "delete",
        }
        action = method_to_action.get(method.upper(), "read")

        # Use path as resource
        resource = path.strip("/").replace("/", ":")
        if not resource:
            resource = "root"

        return resource, action

    async def __call__(
        self,
        scope: dict[str, Any],
        receive: Callable[..., Awaitable[Any]],
        send: Callable[..., Awaitable[Any]],
    ) -> None:
        """Process ASGI request."""
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        path = scope.get("path", "")
        method = scope.get("method", "GET")

        # Check public paths
        if self._middleware.is_public_path(path):
            await self.app(scope, receive, send)
            return

        # Build context
        headers = {}
        for key, value in scope.get("headers", []):
            headers[key.decode()] = value.decode()

        context = {"headers": headers, "path": path, "method": method}

        # Resolve principal
        principal = self._middleware.resolve_principal(context)

        # Map to resource/action
        resource, action = self._resource_mapper(path, method)

        # Check access
        decision = self._middleware.check_access(principal, resource, action)

        if not decision.allowed:
            await self._send_error_response(
                send,
                self._config.denied_status_code,
                decision.reason,
            )
            return

        # Set security context and continue
        if principal:
            with SecurityContext.set_principal(principal):
                await self.app(scope, receive, send)
        else:
            await self.app(scope, receive, send)

    async def _send_error_response(
        self,
        send: Callable[..., Awaitable[Any]],
        status_code: int,
        message: str,
    ) -> None:
        """Send error response."""
        import json

        body = json.dumps({"error": "Permission denied", "message": message}).encode()
        await send({
            "type": "http.response.start",
            "status": status_code,
            "headers": [
                (b"content-type", b"application/json"),
                (b"content-length", str(len(body)).encode()),
            ],
        })
        await send({
            "type": "http.response.body",
            "body": body,
        })


# =============================================================================
# WSGI Middleware
# =============================================================================


class WSGIRBACMiddleware:
    """WSGI middleware for RBAC enforcement.

    Example:
        >>> from flask import Flask
        >>> app = Flask(__name__)
        >>> app.wsgi_app = WSGIRBACMiddleware(
        ...     app.wsgi_app,
        ...     engine=policy_engine,
        ...     principal_store=principal_store,
        ... )
    """

    def __init__(
        self,
        app: Any,
        engine: PolicyEngine,
        principal_store: PrincipalStore,
        config: RBACMiddlewareConfig | None = None,
        resource_mapper: Callable[[str, str], tuple[str, str]] | None = None,
    ) -> None:
        self.app = app
        self._middleware = RBACMiddleware(
            engine=engine,
            principal_store=principal_store,
            config=config,
        )
        self._config = config or RBACMiddlewareConfig()
        self._resource_mapper = resource_mapper or self._default_resource_mapper

    def _default_resource_mapper(self, path: str, method: str) -> tuple[str, str]:
        """Default resource/action mapper."""
        method_to_action = {
            "GET": "read",
            "HEAD": "read",
            "POST": "create",
            "PUT": "update",
            "PATCH": "update",
            "DELETE": "delete",
        }
        action = method_to_action.get(method.upper(), "read")
        resource = path.strip("/").replace("/", ":") or "root"
        return resource, action

    def __call__(
        self,
        environ: dict[str, Any],
        start_response: Callable[..., Any],
    ) -> Any:
        """Process WSGI request."""
        path = environ.get("PATH_INFO", "")
        method = environ.get("REQUEST_METHOD", "GET")

        # Check public paths
        if self._middleware.is_public_path(path):
            return self.app(environ, start_response)

        # Build context
        headers = {}
        for key, value in environ.items():
            if key.startswith("HTTP_"):
                header_name = key[5:].replace("_", "-").title()
                headers[header_name] = value

        context = {"headers": headers, "path": path, "method": method}

        # Resolve principal
        principal = self._middleware.resolve_principal(context)

        # Map to resource/action
        resource, action = self._resource_mapper(path, method)

        # Check access
        decision = self._middleware.check_access(principal, resource, action)

        if not decision.allowed:
            return self._error_response(
                start_response,
                self._config.denied_status_code,
                decision.reason,
            )

        # Set security context and continue
        if principal:
            with SecurityContext.set_principal(principal):
                return self.app(environ, start_response)
        else:
            return self.app(environ, start_response)

    def _error_response(
        self,
        start_response: Callable[..., Any],
        status_code: int,
        message: str,
    ) -> list[bytes]:
        """Generate error response."""
        import json

        body = json.dumps({"error": "Permission denied", "message": message})
        status = f"{status_code} Forbidden"
        headers = [
            ("Content-Type", "application/json"),
            ("Content-Length", str(len(body))),
        ]
        start_response(status, headers)
        return [body.encode()]


# =============================================================================
# Decorators
# =============================================================================


def require_permission(
    resource: str | ResourceType,
    action: str | PermissionAction,
    resource_attributes: dict[str, Any] | None = None,
    engine: PolicyEngine | None = None,
) -> Callable[[F], F]:
    """Decorator that requires a specific permission.

    Example:
        >>> @require_permission("dataset", "read")
        ... def get_dataset(dataset_id: str):
        ...     return load_dataset(dataset_id)
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            principal = SecurityContext.get_current_principal()
            if principal is None:
                raise PermissionDeniedError("No authenticated principal")

            # Use provided engine or get from somewhere
            policy_engine = engine
            if policy_engine is None:
                policy_engine = _get_default_engine()

            # Normalize resource and action
            res = resource.value if isinstance(resource, ResourceType) else resource
            act = action.value if isinstance(action, PermissionAction) else action

            # Check permission
            decision = policy_engine.check(
                principal=principal,
                resource=res,
                action=act,
                resource_attributes=resource_attributes,
            )

            if not decision.allowed:
                raise PermissionDeniedError(
                    decision.reason,
                    principal_id=principal.id,
                    resource=res,
                    action=act,
                )

            return func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def require_permission_async(
    resource: str | ResourceType,
    action: str | PermissionAction,
    resource_attributes: dict[str, Any] | None = None,
    engine: PolicyEngine | None = None,
) -> Callable[[AsyncF], AsyncF]:
    """Async decorator that requires a specific permission.

    Example:
        >>> @require_permission_async("dataset", "read")
        ... async def get_dataset(dataset_id: str):
        ...     return await load_dataset(dataset_id)
    """

    def decorator(func: AsyncF) -> AsyncF:
        @functools.wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            principal = SecurityContext.get_current_principal()
            if principal is None:
                raise PermissionDeniedError("No authenticated principal")

            policy_engine = engine
            if policy_engine is None:
                policy_engine = _get_default_engine()

            res = resource.value if isinstance(resource, ResourceType) else resource
            act = action.value if isinstance(action, PermissionAction) else action

            decision = policy_engine.check(
                principal=principal,
                resource=res,
                action=act,
                resource_attributes=resource_attributes,
            )

            if not decision.allowed:
                raise PermissionDeniedError(
                    decision.reason,
                    principal_id=principal.id,
                    resource=res,
                    action=act,
                )

            return await func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def require_role(
    role: str | set[str],
    require_all: bool = False,
) -> Callable[[F], F]:
    """Decorator that requires specific role(s).

    Example:
        >>> @require_role("admin")
        ... def admin_function():
        ...     pass

        >>> @require_role({"editor", "admin"}, require_all=False)
        ... def editor_or_admin():
        ...     pass
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            principal = SecurityContext.get_current_principal()
            if principal is None:
                raise PermissionDeniedError("No authenticated principal")

            required_roles = {role} if isinstance(role, str) else role

            if require_all:
                # Must have all roles
                if not required_roles.issubset(principal.roles):
                    missing = required_roles - principal.roles
                    raise PermissionDeniedError(
                        f"Missing required roles: {missing}",
                        principal_id=principal.id,
                    )
            else:
                # Must have at least one role
                if not required_roles.intersection(principal.roles):
                    raise PermissionDeniedError(
                        f"Requires one of roles: {required_roles}",
                        principal_id=principal.id,
                    )

            return func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def require_role_async(
    role: str | set[str],
    require_all: bool = False,
) -> Callable[[AsyncF], AsyncF]:
    """Async decorator that requires specific role(s)."""

    def decorator(func: AsyncF) -> AsyncF:
        @functools.wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            principal = SecurityContext.get_current_principal()
            if principal is None:
                raise PermissionDeniedError("No authenticated principal")

            required_roles = {role} if isinstance(role, str) else role

            if require_all:
                if not required_roles.issubset(principal.roles):
                    missing = required_roles - principal.roles
                    raise PermissionDeniedError(
                        f"Missing required roles: {missing}",
                        principal_id=principal.id,
                    )
            else:
                if not required_roles.intersection(principal.roles):
                    raise PermissionDeniedError(
                        f"Requires one of roles: {required_roles}",
                        principal_id=principal.id,
                    )

            return await func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def with_principal(principal: Principal) -> Callable[[F], F]:
    """Decorator that sets principal context for a function.

    Example:
        >>> @with_principal(admin_principal)
        ... def admin_operation():
        ...     pass
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with SecurityContext.set_principal(principal):
                return func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


def with_principal_async(principal: Principal) -> Callable[[AsyncF], AsyncF]:
    """Async decorator that sets principal context."""

    def decorator(func: AsyncF) -> AsyncF:
        @functools.wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            with SecurityContext.set_principal(principal):
                return await func(*args, **kwargs)

        return wrapper  # type: ignore

    return decorator


# =============================================================================
# Flask Integration
# =============================================================================


def create_flask_rbac_middleware(
    engine: PolicyEngine,
    principal_store: PrincipalStore,
    config: RBACMiddlewareConfig | None = None,
    resource_mapper: Callable[[str, str], tuple[str, str]] | None = None,
) -> Callable[[Any], Any]:
    """Create Flask before_request handler for RBAC.

    Example:
        >>> from flask import Flask
        >>> app = Flask(__name__)
        >>> handler = create_flask_rbac_middleware(engine, principal_store)
        >>> app.before_request(handler)
    """
    middleware = RBACMiddleware(
        engine=engine,
        principal_store=principal_store,
        config=config,
    )
    _config = config or RBACMiddlewareConfig()

    def _default_mapper(path: str, method: str) -> tuple[str, str]:
        method_to_action = {
            "GET": "read", "POST": "create", "PUT": "update",
            "PATCH": "update", "DELETE": "delete",
        }
        action = method_to_action.get(method.upper(), "read")
        resource = path.strip("/").replace("/", ":") or "root"
        return resource, action

    mapper = resource_mapper or _default_mapper

    def before_request() -> Any:
        from flask import request, g, abort

        path = request.path
        method = request.method

        # Check public paths
        if middleware.is_public_path(path):
            return None

        context = {
            "headers": dict(request.headers),
            "path": path,
            "method": method,
        }

        # Resolve principal
        principal = middleware.resolve_principal(context)

        # Map to resource/action
        resource, action = mapper(path, method)

        # Check access
        decision = middleware.check_access(principal, resource, action)

        if not decision.allowed:
            if _config.raise_on_denied:
                abort(_config.denied_status_code, decision.reason)
            return None

        # Store principal in g
        g.principal = principal
        if principal:
            g.security_context = SecurityContext.set_principal(principal).__enter__()
        else:
            g.security_context = None

        return None

    return before_request


def create_flask_rbac_teardown(
) -> Callable[[Any], None]:
    """Create Flask teardown handler for RBAC cleanup."""

    def teardown_request(exception: Any = None) -> None:
        from flask import g

        if hasattr(g, "security_context") and g.security_context:
            g.security_context.__exit__(None, None, None)

    return teardown_request


# =============================================================================
# FastAPI Integration
# =============================================================================


def create_fastapi_rbac_dependency(
    engine: PolicyEngine,
    principal_store: PrincipalStore,
    config: RBACMiddlewareConfig | None = None,
) -> Callable[..., Principal]:
    """Create FastAPI dependency for principal injection.

    Example:
        >>> from fastapi import FastAPI, Depends
        >>> app = FastAPI()
        >>> get_principal = create_fastapi_rbac_dependency(engine, principal_store)
        >>>
        >>> @app.get("/data")
        >>> async def get_data(principal: Principal = Depends(get_principal)):
        ...     return {"user": principal.name}
    """
    middleware = RBACMiddleware(
        engine=engine,
        principal_store=principal_store,
        config=config,
    )
    _config = config or RBACMiddlewareConfig()

    async def get_principal(request: Any) -> Principal:
        from fastapi import HTTPException

        context = {
            "headers": dict(request.headers),
            "path": request.url.path,
            "method": request.method,
        }

        principal = middleware.resolve_principal(context)

        if principal is None:
            if _config.deny_anonymous:
                raise HTTPException(
                    status_code=401,
                    detail="Authentication required",
                )
            principal = Principal.anonymous()

        return principal

    return get_principal


def create_fastapi_permission_dependency(
    engine: PolicyEngine,
    principal_store: PrincipalStore,
    resource: str,
    action: str,
    config: RBACMiddlewareConfig | None = None,
) -> Callable[..., Principal]:
    """Create FastAPI dependency that requires a specific permission.

    Example:
        >>> require_read = create_fastapi_permission_dependency(
        ...     engine, principal_store, "dataset", "read"
        ... )
        >>>
        >>> @app.get("/datasets")
        >>> async def list_datasets(principal: Principal = Depends(require_read)):
        ...     return get_datasets()
    """
    middleware = RBACMiddleware(
        engine=engine,
        principal_store=principal_store,
        config=config,
    )
    _config = config or RBACMiddlewareConfig()

    async def check_permission(request: Any) -> Principal:
        from fastapi import HTTPException

        context = {
            "headers": dict(request.headers),
            "path": request.url.path,
            "method": request.method,
        }

        principal = middleware.resolve_principal(context)

        if principal is None:
            if _config.deny_anonymous:
                raise HTTPException(status_code=401, detail="Authentication required")
            principal = Principal.anonymous()

        decision = engine.check(principal, resource, action)

        if not decision.allowed:
            raise HTTPException(
                status_code=_config.denied_status_code,
                detail=decision.reason,
            )

        return principal

    return check_permission


# =============================================================================
# Default Engine Management
# =============================================================================


_default_engine: PolicyEngine | None = None


def set_default_engine(engine: PolicyEngine) -> None:
    """Set the default policy engine."""
    global _default_engine
    _default_engine = engine


def _get_default_engine() -> PolicyEngine:
    """Get the default policy engine."""
    if _default_engine is None:
        raise RuntimeError("No default policy engine configured")
    return _default_engine



"""Base classes for external service integration tests.

This module provides the abstract base classes and protocols that define
the interface for external service integration testing.

Architecture Overview:

    ServiceProvider (Protocol)
        - Abstracts service provisioning (Docker, Local, Cloud, Mock)
        - Handles lifecycle: start, stop, health check
        - Provider-specific configuration

    ExternalServiceBackend (ABC)
        - Platform-specific client creation
        - Connection management with retry logic
        - Metrics collection and cost tracking

    ServiceRegistry (Singleton)
        - Central registry of available services
        - Provider discovery and caching
        - Service dependency resolution

Design Decisions:
    1. Provider Pattern: Same test can run against different providers
    2. Health Check Protocol: Uniform health checking across all services
    3. Context Managers: Automatic cleanup on test completion
    4. Cost Tracking: Cloud services track estimated costs
"""

from __future__ import annotations

import abc
import atexit
import logging
import os
import threading
import time
import uuid
from collections.abc import Mapping
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum, auto
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Iterator,
    Protocol,
    TypeVar,
    runtime_checkable,
)

if TYPE_CHECKING:
    from docker import DockerClient

logger = logging.getLogger(__name__)


# =============================================================================
# Enums and Constants
# =============================================================================


class ServiceStatus(Enum):
    """Status of an external service."""

    UNKNOWN = auto()
    STARTING = auto()
    RUNNING = auto()
    HEALTHY = auto()
    UNHEALTHY = auto()
    STOPPING = auto()
    STOPPED = auto()
    FAILED = auto()
    UNAVAILABLE = auto()


class ServiceCategory(Enum):
    """Categories of external services."""

    CACHE = "cache"  # Redis, Memcached
    MESSAGE_QUEUE = "message_queue"  # Kafka, RabbitMQ
    SEARCH = "search"  # Elasticsearch
    LOGGING = "logging"  # Loki, Fluentd
    SECRETS = "secrets"  # Vault, Cloud KMS
    STORAGE = "storage"  # S3, GCS, MinIO
    DATABASE = "database"  # PostgreSQL, MySQL
    TMS = "tms"  # Translation Management Systems
    MONITORING = "monitoring"  # Prometheus


class ProviderType(Enum):
    """Types of service providers."""

    DOCKER = "docker"
    LOCAL = "local"
    CLOUD = "cloud"
    MOCK = "mock"


# =============================================================================
# Configuration
# =============================================================================


@dataclass
class ServiceConfig:
    """Configuration for an external service.

    Attributes:
        name: Service name (e.g., "redis", "elasticsearch")
        category: Service category
        provider: Preferred provider type
        host: Service host (override)
        port: Service port (override)
        timeout_seconds: Connection timeout
        retry_attempts: Number of retry attempts
        retry_delay_seconds: Delay between retries
        health_check_interval: Health check interval in seconds
        cleanup_on_exit: Whether to clean up on process exit
        environment: Additional environment variables
        labels: Metadata labels for the service
    """

    name: str
    category: ServiceCategory
    provider: ProviderType = ProviderType.DOCKER
    host: str | None = None
    port: int | None = None
    timeout_seconds: int = 30
    retry_attempts: int = 3
    retry_delay_seconds: float = 1.0
    health_check_interval: float = 5.0
    cleanup_on_exit: bool = True
    environment: dict[str, str] = field(default_factory=dict)
    labels: dict[str, str] = field(default_factory=dict)

    # Cost tracking for cloud services
    max_cost_usd: float = 1.0
    track_cost: bool = True

    @classmethod
    def from_env(cls, name: str, category: ServiceCategory) -> "ServiceConfig":
        """Create configuration from environment variables.

        Environment variable naming convention:
            TRUTHOUND_TEST_{NAME}_{SETTING}

        Example:
            TRUTHOUND_TEST_REDIS_HOST=localhost
            TRUTHOUND_TEST_REDIS_PORT=6379
        """
        prefix = f"TRUTHOUND_TEST_{name.upper()}"

        provider_str = os.getenv(f"{prefix}_PROVIDER", "docker")
        provider = ProviderType(provider_str.lower())

        return cls(
            name=name,
            category=category,
            provider=provider,
            host=os.getenv(f"{prefix}_HOST"),
            port=int(os.getenv(f"{prefix}_PORT", "0")) or None,
            timeout_seconds=int(os.getenv(f"{prefix}_TIMEOUT", "30")),
            retry_attempts=int(os.getenv(f"{prefix}_RETRIES", "3")),
            cleanup_on_exit=os.getenv(f"{prefix}_CLEANUP", "true").lower() == "true",
            max_cost_usd=float(os.getenv(f"{prefix}_MAX_COST", "1.0")),
        )


@dataclass
class ServiceMetrics:
    """Metrics collected for a service during testing.

    Attributes:
        operation_count: Number of operations performed
        total_duration_seconds: Total time spent on operations
        error_count: Number of errors encountered
        bytes_transferred: Total bytes transferred (if applicable)
        estimated_cost_usd: Estimated cost for cloud services
        operations: List of operation details
    """

    operation_count: int = 0
    total_duration_seconds: float = 0.0
    error_count: int = 0
    bytes_transferred: int = 0
    estimated_cost_usd: float = 0.0
    operations: list[dict[str, Any]] = field(default_factory=list)

    def record_operation(
        self,
        operation: str,
        duration_seconds: float,
        success: bool = True,
        bytes_count: int = 0,
        cost_usd: float = 0.0,
        error: str | None = None,
        **metadata: Any,
    ) -> None:
        """Record an operation."""
        self.operation_count += 1
        self.total_duration_seconds += duration_seconds
        self.bytes_transferred += bytes_count
        self.estimated_cost_usd += cost_usd

        if not success:
            self.error_count += 1

        self.operations.append(
            {
                "operation": operation,
                "duration_seconds": duration_seconds,
                "success": success,
                "bytes_count": bytes_count,
                "cost_usd": cost_usd,
                "error": error,
                "timestamp": datetime.utcnow().isoformat(),
                **metadata,
            }
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            "operation_count": self.operation_count,
            "total_duration_seconds": self.total_duration_seconds,
            "error_count": self.error_count,
            "bytes_transferred": self.bytes_transferred,
            "estimated_cost_usd": self.estimated_cost_usd,
            "avg_operation_duration": (
                self.total_duration_seconds / self.operation_count
                if self.operation_count > 0
                else 0
            ),
            "error_rate": (
                self.error_count / self.operation_count
                if self.operation_count > 0
                else 0
            ),
        }


# =============================================================================
# Health Check Protocol
# =============================================================================


@dataclass
class HealthCheckResult:
    """Result of a health check.

    Attributes:
        healthy: Whether the service is healthy
        message: Status message
        latency_ms: Health check latency in milliseconds
        details: Additional health details
        timestamp: When the check was performed
    """

    healthy: bool
    message: str = ""
    latency_ms: float = 0.0
    details: dict[str, Any] = field(default_factory=dict)
    timestamp: datetime = field(default_factory=datetime.utcnow)

    @classmethod
    def success(cls, message: str = "OK", latency_ms: float = 0.0, **details: Any) -> "HealthCheckResult":
        """Create a successful health check result."""
        return cls(healthy=True, message=message, latency_ms=latency_ms, details=details)

    @classmethod
    def failure(cls, message: str, latency_ms: float = 0.0, **details: Any) -> "HealthCheckResult":
        """Create a failed health check result."""
        return cls(healthy=False, message=message, latency_ms=latency_ms, details=details)


@runtime_checkable
class HealthCheckable(Protocol):
    """Protocol for services that support health checking."""

    def health_check(self) -> HealthCheckResult:
        """Perform a health check.

        Returns:
            HealthCheckResult with status and details.
        """
        ...


# =============================================================================
# Service Provider Protocol
# =============================================================================


ConfigT = TypeVar("ConfigT", bound=ServiceConfig)
ClientT = TypeVar("ClientT")


@runtime_checkable
class ServiceProvider(Protocol[ConfigT]):
    """Protocol for service providers.

    Service providers handle the lifecycle of external services,
    abstracting the difference between Docker, local, cloud, and mock services.
    """

    provider_type: ClassVar[ProviderType]

    def is_available(self) -> bool:
        """Check if this provider is available."""
        ...

    def start_service(self, config: ConfigT) -> dict[str, Any]:
        """Start the service.

        Returns:
            Connection info dictionary with host, port, etc.
        """
        ...

    def stop_service(self, config: ConfigT) -> None:
        """Stop the service."""
        ...

    def get_connection_info(self, config: ConfigT) -> dict[str, Any]:
        """Get connection information for the service.

        Returns:
            Dictionary with host, port, and other connection details.
        """
        ...

    def health_check(self, config: ConfigT) -> HealthCheckResult:
        """Check if the service is healthy."""
        ...


# =============================================================================
# Abstract Service Backend
# =============================================================================


class ExternalServiceBackend(abc.ABC, Generic[ConfigT, ClientT]):
    """Abstract base class for external service test backends.

    This class defines the interface that all external service backends
    must implement. It provides:
    - Connection management with retry logic
    - Health checking and wait-for-ready
    - Metrics collection
    - Automatic cleanup

    Type Parameters:
        ConfigT: Service configuration type
        ClientT: Client type for the service

    Example:
        >>> class RedisBackend(ExternalServiceBackend[RedisConfig, Redis]):
        ...     service_name = "redis"
        ...     service_category = ServiceCategory.CACHE
        ...
        ...     def _create_client(self) -> Redis:
        ...         return Redis(host=self.host, port=self.port)
    """

    # Class-level attributes (must be overridden)
    service_name: ClassVar[str] = "unknown"
    service_category: ClassVar[ServiceCategory] = ServiceCategory.CACHE
    default_port: ClassVar[int] = 0
    default_image: ClassVar[str] = ""

    def __init__(
        self,
        config: ConfigT,
        provider: ServiceProvider[ConfigT] | None = None,
    ) -> None:
        """Initialize the backend.

        Args:
            config: Service configuration
            provider: Optional service provider (auto-detected if not provided)
        """
        self.config = config
        self._provider = provider
        self._client: ClientT | None = None
        self._status = ServiceStatus.UNKNOWN
        self._metrics = ServiceMetrics()
        self._connection_info: dict[str, Any] = {}
        self._instance_id = uuid.uuid4().hex[:8]

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------

    @property
    def status(self) -> ServiceStatus:
        """Get current service status."""
        return self._status

    @property
    def is_running(self) -> bool:
        """Check if service is running."""
        return self._status in (ServiceStatus.RUNNING, ServiceStatus.HEALTHY)

    @property
    def is_healthy(self) -> bool:
        """Check if service is healthy."""
        return self._status == ServiceStatus.HEALTHY

    @property
    def host(self) -> str:
        """Get service host."""
        return self._connection_info.get("host", self.config.host or "localhost")

    @property
    def port(self) -> int:
        """Get service port."""
        return self._connection_info.get("port", self.config.port or self.default_port)

    @property
    def client(self) -> ClientT:
        """Get the service client.

        Raises:
            RuntimeError: If not connected.
        """
        if self._client is None:
            raise RuntimeError(f"{self.service_name} client not connected")
        return self._client

    @property
    def metrics(self) -> ServiceMetrics:
        """Get collected metrics."""
        return self._metrics

    # -------------------------------------------------------------------------
    # Provider Management
    # -------------------------------------------------------------------------

    def get_provider(self) -> ServiceProvider[ConfigT]:
        """Get or create the service provider."""
        if self._provider is not None:
            return self._provider

        # Auto-detect provider
        from tests.integration.external.providers import get_provider

        self._provider = get_provider(self.config.provider, self.config)
        return self._provider

    # -------------------------------------------------------------------------
    # Lifecycle Management
    # -------------------------------------------------------------------------

    def start(self) -> bool:
        """Start the service and connect.

        Returns:
            True if successful, False otherwise.
        """
        if self.is_running:
            return True

        try:
            self._status = ServiceStatus.STARTING
            logger.info(f"[{self.service_name}] Starting service...")

            # Start the service via provider
            provider = self.get_provider()
            self._connection_info = provider.start_service(self.config)

            self._status = ServiceStatus.RUNNING
            logger.info(
                f"[{self.service_name}] Service started at "
                f"{self.host}:{self.port}"
            )

            # Connect client
            return self.connect()

        except Exception as e:
            self._status = ServiceStatus.FAILED
            logger.error(f"[{self.service_name}] Failed to start: {e}")
            return False

    def stop(self) -> None:
        """Stop the service."""
        if self._status in (ServiceStatus.STOPPED, ServiceStatus.UNKNOWN):
            return

        try:
            self._status = ServiceStatus.STOPPING
            logger.info(f"[{self.service_name}] Stopping service...")

            # Disconnect client
            self.disconnect()

            # Stop service via provider
            if self._provider is not None:
                self._provider.stop_service(self.config)

            self._status = ServiceStatus.STOPPED
            logger.info(f"[{self.service_name}] Service stopped")

        except Exception as e:
            logger.warning(f"[{self.service_name}] Error stopping: {e}")
            self._status = ServiceStatus.STOPPED

    def connect(self) -> bool:
        """Connect to the service.

        Returns:
            True if connected successfully.
        """
        if self._client is not None:
            return True

        try:
            self._client = self._create_client()
            logger.info(f"[{self.service_name}] Connected")
            return True

        except Exception as e:
            logger.error(f"[{self.service_name}] Connection failed: {e}")
            return False

    def disconnect(self) -> None:
        """Disconnect from the service."""
        if self._client is not None:
            try:
                self._close_client()
            except Exception as e:
                logger.warning(f"[{self.service_name}] Error disconnecting: {e}")
            finally:
                self._client = None
                logger.info(f"[{self.service_name}] Disconnected")

    def wait_until_ready(
        self,
        timeout_seconds: float | None = None,
        check_interval: float | None = None,
    ) -> bool:
        """Wait until the service is ready.

        Args:
            timeout_seconds: Maximum time to wait
            check_interval: Time between checks

        Returns:
            True if service became ready, False if timeout.
        """
        timeout = timeout_seconds or self.config.timeout_seconds
        interval = check_interval or self.config.health_check_interval

        start_time = time.time()
        while time.time() - start_time < timeout:
            result = self.health_check()
            if result.healthy:
                self._status = ServiceStatus.HEALTHY
                logger.info(
                    f"[{self.service_name}] Service ready "
                    f"(latency: {result.latency_ms:.1f}ms)"
                )
                return True

            time.sleep(interval)

        logger.warning(f"[{self.service_name}] Timeout waiting for service")
        self._status = ServiceStatus.UNHEALTHY
        return False

    # -------------------------------------------------------------------------
    # Health Checking
    # -------------------------------------------------------------------------

    def health_check(self) -> HealthCheckResult:
        """Perform a health check.

        Returns:
            HealthCheckResult with status and details.
        """
        start_time = time.time()
        try:
            result = self._perform_health_check()
            result.latency_ms = (time.time() - start_time) * 1000
            return result
        except Exception as e:
            latency_ms = (time.time() - start_time) * 1000
            return HealthCheckResult.failure(str(e), latency_ms=latency_ms)

    # -------------------------------------------------------------------------
    # Abstract Methods
    # -------------------------------------------------------------------------

    @abc.abstractmethod
    def _create_client(self) -> ClientT:
        """Create the service client.

        Returns:
            Connected client instance.
        """
        ...

    @abc.abstractmethod
    def _close_client(self) -> None:
        """Close the service client."""
        ...

    @abc.abstractmethod
    def _perform_health_check(self) -> HealthCheckResult:
        """Perform service-specific health check.

        Returns:
            HealthCheckResult with status.
        """
        ...

    # -------------------------------------------------------------------------
    # Context Manager
    # -------------------------------------------------------------------------

    @contextmanager
    def session(self) -> Iterator["ExternalServiceBackend[ConfigT, ClientT]"]:
        """Context manager for a test session.

        Automatically starts and stops the service.

        Example:
            >>> with backend.session() as svc:
            ...     svc.client.set("key", "value")
        """
        try:
            self.start()
            self.wait_until_ready()
            yield self
        finally:
            if self.config.cleanup_on_exit:
                self.stop()

    def __enter__(self) -> "ExternalServiceBackend[ConfigT, ClientT]":
        """Context manager entry."""
        self.start()
        self.wait_until_ready()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Context manager exit."""
        if self.config.cleanup_on_exit or exc_type is None:
            self.stop()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"service={self.service_name!r}, "
            f"status={self._status.name}, "
            f"host={self.host}:{self.port})"
        )


# =============================================================================
# Service Registry
# =============================================================================


class ServiceRegistry:
    """Central registry for external services.

    Provides singleton access to service backends and handles
    service discovery, caching, and dependency resolution.

    Example:
        >>> registry = ServiceRegistry.instance()
        >>> redis = registry.get_backend("redis")
        >>> if redis.is_available():
        ...     redis.start()
    """

    _instance: ClassVar["ServiceRegistry | None"] = None
    _lock: ClassVar[threading.Lock] = threading.Lock()

    def __init__(self) -> None:
        """Initialize the registry."""
        self._backends: dict[str, ExternalServiceBackend[Any, Any]] = {}
        self._providers: dict[ProviderType, ServiceProvider[Any]] = {}
        self._started_services: list[str] = []

        # Register cleanup on exit
        atexit.register(self._cleanup)

    @classmethod
    def instance(cls) -> "ServiceRegistry":
        """Get the singleton instance."""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = cls()
        return cls._instance

    @classmethod
    def reset(cls) -> None:
        """Reset the registry (for testing)."""
        if cls._instance is not None:
            cls._instance._cleanup()
            cls._instance = None

    def register_backend(
        self,
        name: str,
        backend: ExternalServiceBackend[Any, Any],
    ) -> None:
        """Register a service backend."""
        self._backends[name] = backend
        logger.debug(f"Registered backend: {name}")

    def get_backend(self, name: str) -> ExternalServiceBackend[Any, Any] | None:
        """Get a registered backend."""
        return self._backends.get(name)

    def get_or_create_backend(
        self,
        name: str,
        factory: type[ExternalServiceBackend[Any, Any]],
        config: ServiceConfig,
    ) -> ExternalServiceBackend[Any, Any]:
        """Get or create a backend."""
        if name not in self._backends:
            backend = factory(config)
            self.register_backend(name, backend)
        return self._backends[name]

    def list_backends(self) -> list[str]:
        """List all registered backend names."""
        return list(self._backends.keys())

    def start_service(self, name: str) -> bool:
        """Start a service by name."""
        backend = self.get_backend(name)
        if backend is None:
            logger.error(f"Unknown service: {name}")
            return False

        if backend.start():
            self._started_services.append(name)
            return True
        return False

    def stop_service(self, name: str) -> None:
        """Stop a service by name."""
        backend = self.get_backend(name)
        if backend is not None:
            backend.stop()
            if name in self._started_services:
                self._started_services.remove(name)

    def stop_all(self) -> None:
        """Stop all started services."""
        for name in reversed(self._started_services[:]):
            self.stop_service(name)

    def _cleanup(self) -> None:
        """Cleanup on exit."""
        logger.info("Cleaning up external services...")
        self.stop_all()



"""Timezone validators."""

from typing import Any

import polars as pl

from truthound.types import Severity
from truthound.validators.base import ValidationIssue, Validator
from truthound.validators.registry import register_validator


@register_validator
class TimezoneValidator(Validator):
    """Validates timezone consistency in datetime columns."""

    name = "timezone"
    category = "datetime"

    def __init__(
        self,
        expected_timezone: str | None = None,
        require_timezone: bool = True,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.expected_timezone = expected_timezone
        self.require_timezone = require_timezone

    def validate(self, lf: pl.LazyFrame) -> list[ValidationIssue]:
        issues: list[ValidationIssue] = []
        schema = lf.collect_schema()
        columns = self._get_target_columns(lf)

        for col in columns:
            dtype = schema[col]

            # Only check Datetime columns
            if not isinstance(dtype, pl.Datetime):
                continue

            tz = dtype.time_zone

            if self.require_timezone and tz is None:
                issues.append(
                    ValidationIssue(
                        column=col,
                        issue_type="missing_timezone",
                        count=1,
                        severity=Severity.MEDIUM,
                        details="Datetime column has no timezone",
                        expected="timezone-aware",
                        actual="timezone-naive",
                    )
                )

            if self.expected_timezone and tz != self.expected_timezone:
                issues.append(
                    ValidationIssue(
                        column=col,
                        issue_type="wrong_timezone",
                        count=1,
                        severity=Severity.MEDIUM,
                        details=f"Expected {self.expected_timezone}, got {tz}",
                        expected=self.expected_timezone,
                        actual=tz,
                    )
                )

        return issues
